import re
import json
import os

try:
    import hvac
except ImportError as err:
    raise ImportError(
        "Using `prefect.tasks.secrets.vault_secret` requires hvac to be installed."
    ) from err

import prefect
from prefect.tasks.secrets import SecretBase, PrefectSecret
from prefect.utilities.tasks import defaults_from_attrs


class VaultSecret(SecretBase):
    """
    Vault Secrets Task.  This task retrieves a secret using the Hashicorp Vault service.

    Client configuration:
        - vault_addr: set via env var.  Both `VAULT_ADDR` and `vault_addr` supported

    Args:
        - name (str): secret name defined by the Vault secret path "<mount-point>/<path>"
        - vault_credentials_secret (str, optional): name of the PrefectSecret containing Vault
            credentials.
                Defaults to a PrefectSecret named `VAULT_CREDENTIALS`.
                Supported vault client authentication methods:
                * token { 'VAULT_TOKEN: '<token>' }
                * appRole: { 'VAULT_ROLE_ID': '<role-id>',
                             'VAULT_SECRET_ID': '<secret-id>' }
        - **kwargs (Any, optional): additional keyword args passed to the Task constructor

    Raises:
        - ValueError: if a `result` keyword is passed
    """

    def __init__(
        self, name: str, vault_credentials_secret: str = "VAULT_CREDENTIALS", **kwargs
    ):
        self.vault_credentials_secret = vault_credentials_secret
        kwargs["name"] = name
        super().__init__(**kwargs)

    def _get_vault_secret(self, name: str):
        """
        Connect to vault service and retrieve a secret value
        Arguments:
            - name (str): name of the secret to be retrieved using "<mount_point>/<path>"

        Returns:
            - secret (dict): a dict of the secret key/value items retrieved

        Raises:
            - ValueError: unable to configure the vault client or unable to parse the provided
                secret name into a "<mount_point>/<path>" pattern
            - RuntimeError: unable to authenticate or attempting an unsupported auth method or
                or provided token/role not authorised to access the the secret
            - KeyError: unable to lookup secret in vault using the secret path
        """
        self.logger.debug(f"Looking up vault path: {name}")
        client = hvac.Client()
        # get vault address url
        vault_url = os.getenv("VAULT_ADDR") or os.getenv("vault_addr")
        if not vault_url:
            raise ValueError(
                "VAULT_ADDR url var not found. "
                'Either "VAULT_ADDR" or "vault_addr" env var required.'
            )
        client.url = vault_url
        self.logger.debug(f"Vault addr set to: {client.url}")

        # get vault auth credentials from the PrefectSecret
        vault_creds = PrefectSecret(self.vault_credentials_secret).run()
        if "VAULT_TOKEN" in vault_creds.keys():
            client.token = vault_creds["VAULT_TOKEN"]
        elif (
            "VAULT_ROLE_ID" in vault_creds.keys()
            and "VAULT_SECRET_ID" in vault_creds.keys()
        ):
            client.auth_approle(
                vault_creds["VAULT_ROLE_ID"], vault_creds["VAULT_SECRET_ID"]
            )
        else:
            raise ValueError(
                "Unable to authenticate with vault service.  "
                "Supported methods: token, appRole"
            )
        if not client.is_authenticated():
            raise RuntimeError(
                "Unable to autheticate with vault using supplied credentials"
            )
        self.logger.debug("Passed vault authentication check")
        # regex to parse path into 2 named parts: <mount_point>/<path>
        secret_path_re = r"^(?P<mount_point>[^/]+)/(?P<path>.+)$"
        m = re.fullmatch(secret_path_re, name)
        if m is None:
            raise ValueError(
                f'Invalid secret path: {name}.  Expected: "<mount_point>/<path>"'
            )
        vault_path = m.groupdict()
        value = ""
        try:
            vault_secret = client.secrets.kv.v2.read_secret_version(
                path=vault_path["path"], mount_point=vault_path["mount_point"]
            )
            value = vault_secret["data"]["data"]
        except hvac.exceptions.InvalidPath as exc:
            raise KeyError(f"Secret not found: {name}") from exc
        except hvac.exceptions.Forbidden as exc:
            raise RuntimeError(f"Access forbidden: {name}") from exc
        return value

    @defaults_from_attrs("name")
    def run(self, name: str = None):
        """
        The run method for VaultSecret Task.

        Args:
            - name (str, optional): The secret name defined by the Vault secret
                path "<mount-point>/<path>".  Defaults to the name provided at initialization.

        Returns:
            - Any: the underlying value of the Vault Secret

        Raises:
            - ValueError: if a `result` keyword is passed, or if called within a flow building context
                or if unable to find VAULT_ADDR or vault_addr environment variable
            - RuntimeError: if `use_local_secrets=False` and the Client is not authorised to read
                the secret
        """
        if isinstance(prefect.context.get("flow"), prefect.core.flow.Flow):
            raise ValueError(
                "Secrets should only be retrieved during a Flow run, "
                "not while building a Flow."
            )

        value = None
        try:
            # lookup the secret value in Vault
            value = self._get_vault_secret(name)
            return json.loads(value)
        except (json.JSONDecodeError, TypeError):
            return value



from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

import sys
from pyarrow import flight
import pyarrow as pa
from typing import Any, Dict


class DremioClientAuthMiddlewareFactory(flight.ClientMiddlewareFactory):
    """
    A factory that creates DremioClientAuthMiddleware(s).

    Args:

    Returns:

    """

    def __init__(self):
        self.call_credential = []

    def start_call(self, info):
        return DremioClientAuthMiddleware(self)

    def set_call_credential(self, call_credential):
        self.call_credential = call_credential


class DremioClientAuthMiddleware(flight.ClientMiddleware):
    """
    A ClientMiddleware that extracts the bearer token from
    the authorization header returned by the Dremio
    Flight Server Endpoint.

    Args:
        - factory : ClientHeaderAuthMiddlewareFactory
        The factory to set call credentials if an
        authorization header with bearer token is
        returned by the Dremio server.
    Returns:
        - bearer token
    """

    def __init__(self, factory):
        self.factory = factory

    def received_headers(self, headers):
        auth_header_key = "authorization"
        authorization_header = []
        for key in headers:
            if key.lower() == auth_header_key:
                authorization_header = headers.get(auth_header_key)
        self.factory.set_call_credential(
            [b"authorization", authorization_header[0].encode("utf-8")]
        )


class DremioFetch(Task):
    """
    Task for fetching results of a query using Dremio Query Engine.

    Args:
        - username (str): user name used to authenticate
        - password (str): password used to authenticate
        - hostname (str): Dremio host address
        - flightport (int, optional): port used to connect to FlightClient, defaults to 32010 if not
            provided
        - tls (bool): connect to the server endpoint with an encrypted TLS connection
        - certs (str): path to a certificate
        - query (str, optional): query to execute against Dremio
        - **kwargs (Any, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        username: str = None,
        password: str = None,
        hostname: str = None,
        flightport: int = 32010,
        tls: bool = False,
        certs: str = None,
        query: str = None,
        **kwargs: Any,
    ):
        self.username = username
        self.password = password
        self.hostname = hostname
        self.flightport = flightport
        self.tls = tls
        self.certs = certs
        self.query = query
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "username", "password", "hostname", "flightport", "tls", "certs", "query"
    )
    def run(
        self,
        username: str,
        password: str,
        hostname: str,
        flightport: int,
        tls: bool,
        certs: str,
        query: str,
    ) -> Dict:
        """
        Task run method. Executes a query against Dremio and fetches results.

        Args:
            - username (str): user name used to authenticate
            - password (str): password used to authenticate
            - hostname (str): Dremio host address
            - flightport (int, optional): port used to connect to FlightClient, defaults to 32010 if not
                provided
            - tls (bool): connect to the server endpoint with an encrypted TLS connection
            - certs (str): path to a certificate
            - query (str, optional): query to execute against query engine

        Returns:
            - dict: a dictionary of data returned by Dremio

        Raises:
            - ValueError: if `query` is `None`
        """
        if not query:
            raise ValueError("A query string must be provided")
        scheme = "grpc+tcp"
        connection_args = {}

        if tls:
            # Connect to the server endpoint with an encrypted TLS connection.
            self.logger.debug("Enabling TLS connection")
            scheme = "grpc+tls"
            if certs:
                self.logger.debug("Trusted certificates provided")
                # TLS certificates are provided in a list of connection arguments.
                with open(certs, "rb") as root_certs:
                    connection_args["tls_root_certs"] = root_certs.read()
            else:
                self.logger.error(
                    "Trusted certificates must be provided to establish a TLS connection"
                )
                sys.exit()
        else:
            self.logger.info(
                "You are not using a secure connection. Consider setting tls=True"
            )

        # Two WLM settings can be provided upon initial authneitcation
        # with the Dremio Server Flight Endpoint:
        # - routing-tag
        # - routing queue
        initial_options = flight.FlightCallOptions(
            headers=[
                (b"routing-tag", b"test-routing-tag"),
                (b"routing-queue", b"Low Cost User Queries"),
            ]
        )
        client_auth_middleware = DremioClientAuthMiddlewareFactory()
        client = flight.FlightClient(
            "{}://{}:{}".format(scheme, hostname, flightport),
            middleware=[client_auth_middleware],
            **connection_args,
        )

        # Authenticate with the server endpoint.
        bearer_token = client.authenticate_basic_token(
            username, password, initial_options
        )
        self.logger.debug("Authentication was successful. Token is valid for 30 hours.")

        # Retrieve the schema of the result set.
        options = flight.FlightCallOptions(headers=[bearer_token])

        # Get the FlightInfo message to retrieve the Ticket corresponding
        # to the query result set.
        flight_info = client.get_flight_info(
            flight.FlightDescriptor.for_command(query), options
        )
        self.logger.debug("GetFlightInfo was successful")

        # Retrieve the result set as a stream of Arrow record batches.
        reader = client.do_get(flight_info.endpoints[0].ticket, options)
        self.logger.debug("Reading query results from Dremio")

        # batches of data reduce the number of calls to the server

        batches = []
        while True:
            try:
                batch, _ = reader.read_chunk()
                batches.append(batch)
            except StopIteration:
                break
        data = pa.Table.from_batches(batches)

        return data.to_pydict()



from typing import Iterable, List

from prefect.run_configs.base import RunConfig


class VertexRun(RunConfig):
    """Configure a flow-run to run as a Vertex CustomJob.

    The most common configuration is for the run to use a single machine, and
    prefect will provide a default workerPoolSpec from the machine_type
    argument. But this can be customized by providing the worker_pool_spec arg
    with the content described in [workerPoolSpec][1]. Prefect will always
    provide the container spec for the 0th workerPoolSpec, which is used to
    actually run the flow

    Args:
        - env (dict, optional): Additional environment variables to set.
        - labels (Iterable[str], optional): an iterable of labels to apply to this
            run config. Labels are string identifiers used by Prefect Agents
            for selecting valid flow runs when polling for work
        - image (str, optional): The image to use for this task. If not
            provided, will be either inferred from the flow's storage (if using
            `Docker` storage), or use the default configured on the agent.
        - machine_type (str, optional): The machine type to use for the run,
            which controls the available CPU and memory. See [machine_types][2]
            for valid choices.
        - scheduling (dictionary, optional): The scheduling for the custom job, which
            can be used to set a timeout (maximum run time). See
            [CustomJobSpec][1] for the expected format.
        - service_account (str, optional): Specifies the service account to use
            as the run-as account in vertex. The agent submitting jobs must have
            act-as permission on this run-as account. If unspecified, the AI
            Platform Custom Code Service Agent for the CustomJob's project is
            used.
        - network (str, optional): The full name of the Compute Engine network
            to which the Job should be peered. Private services access must
            already be configured for the network. If left unspecified, the job
            is not peered with any network.
        - worker_pool_specs (list, optional): The full worker pool specs
            for the custom job, which can be used for advanced configuration.
            If provided, it will overwrite the default constructed from the
            machine type arg.


    Examples:

    Use the defaults set on the agent:

    ```python
    flow.run_config = VertexRun()
    ```

    Use the default task definition, but override the image and machine type:

    ```python
    flow.run_config = VertexRun(
        image="example/my-custom-image:latest",
        machine_type="e2-highmem-8",
    )
    ```

    [1]: https://cloud.google.com/vertex-ai/docs/reference/rest/v1/\
CustomJobSpec
    [2]: https://cloud.google.com/vertex-ai/docs/training/\
configure-compute#machine-types

    """

    # TODO ADD GPUS SUPPORT
    def __init__(
        self,
        *,
        env: dict = None,
        labels: Iterable[str] = None,
        image: str = None,
        machine_type: str = "e2-standard-4",
        scheduling: dict = None,
        service_account: str = None,
        network: str = None,
        worker_pool_specs: List[dict] = None,
    ) -> None:
        super().__init__(env=env, labels=labels)
        self.image = image
        self.machine_type = machine_type
        self.scheduling = scheduling
        self.service_account = service_account
        self.network = network
        self.worker_pool_specs = worker_pool_specs



"""
The exception types in this module have been deprecated in favor of `prefect.exceptions`

Users should not be using these directly but we leave these for a version in case they
are being used in try/except clauses
"""

import warnings


class PrefectError(Exception):
    def __init__(self, *args: object) -> None:
        if type(self) == PrefectError:
            warnings.warn(
                "`prefect.utilities.exceptions.PrefectError` has been moved to "
                "`prefect.exceptions.PrefectException` and will be removed in a future "
                "release. Please update your imports.",
                stacklevel=2,
            )
        super().__init__(*args)


class TaskTimeoutError(PrefectError):
    def __init__(self, *args: object) -> None:
        warnings.warn(
            "`prefect.utilities.exceptions.TaskTimeoutError` has been moved to "
            "`prefect.exceptions.TaskTimeoutSignal` and will be removed in a future "
            "release. Please update your imports.",
            stacklevel=2,
        )
        super().__init__(*args)


class ContextError(KeyError, PrefectError):
    def __init__(self, *args: object) -> None:
        warnings.warn(
            "`prefect.utilities.exceptions.ContextError` has been deprecated "
            "and will be removed in a future release.",
            stacklevel=2,
        )
        super().__init__(*args)


class SerializationError(PrefectError):
    def __init__(self, *args: object) -> None:
        warnings.warn(
            "`prefect.utilities.exceptions.SerializationError` has been deprecated "
            "and will be removed in a future release.",
            stacklevel=2,
        )
        super().__init__(*args)


class PrefectWarning(UserWarning):
    def __init__(self, *args: object) -> None:
        if type(self) == PrefectWarning:
            warnings.warn(
                "`prefect.utilities.exceptions.PrefectWarning` has been deprecated "
                "and will be removed in a future release.",
                stacklevel=2,
            )
        super().__init__(*args)


class ClientError(PrefectError):
    def __init__(self, *args: object) -> None:
        if type(self) == ClientError:
            warnings.warn(
                "`prefect.utilities.exceptions.ClientError` has been moved to "
                "`prefect.exceptions.ClientError` and will be removed in a future "
                "release. Please update your imports.",
                stacklevel=2,
            )
        super().__init__(*args)


class VersionLockError(PrefectError):
    def __init__(self, *args: object) -> None:
        warnings.warn(
            "`prefect.utilities.exceptions.VersionLockError` has been moved to "
            "`prefect.exceptions.VersionLockMismatchSignal` and will be removed in a "
            "future release. Please update your imports.",
            stacklevel=2,
        )
        super().__init__(*args)


class AuthorizationError(ClientError):
    def __init__(self, *args: object) -> None:
        if type(self) == AuthorizationError:
            warnings.warn(
                "`prefect.utilities.exceptions.AuthorizationError` has been moved to "
                "`prefect.exceptions.AuthorizationError` and will be removed in a "
                "future release. Please update your imports.",
                stacklevel=2,
            )
        super().__init__(*args)


class StorageError(PrefectError):
    def __init__(self, *args: object) -> None:
        if type(self) == StorageError:
            warnings.warn(
                "`prefect.utilities.exceptions.StorageError` has been moved to "
                "`prefect.exceptions.FlowStorageError` and will be removed in a future "
                "release. Please update your imports.",
                stacklevel=2,
            )
        super().__init__(*args)



import shutil
from pathlib import Path
from typing import Any, Union, List

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class Move(Task):
    """
    Task for moving files or directories within the file system.

    Args:
        - source_path (Union[str, Path], optional): the path to the source directory/file.
        - target_path (Union[str, Path], optional): the path to the target directory/file. Any
            parent directories of `target_path` must already exist.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        source_path: Union[str, Path] = "",
        target_path: Union[str, Path] = "",
        **kwargs: Any,
    ):
        self.source_path = source_path
        self.target_path = target_path
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "source_path",
        "target_path",
    )
    def run(
        self,
        source_path: Union[str, Path] = "",
        target_path: Union[str, Path] = "",
    ) -> Path:
        """
        Task run method.

        Args:
            - source_path (Union[str, Path], optional): the path to the source directory/file.
            - target_path (Union[str, Path], optional): the path to the target directory/file. Any
                parent directories of `target_path` must already exist.

        Returns:
            - Path: resulting path of the moved file / directory
        """
        if not source_path:
            raise ValueError("No `source_path` provided.")
        if not target_path:
            raise ValueError("No `target_path` provided.")

        source_path = Path(source_path)
        target_path = Path(target_path)

        if not source_path.exists():
            raise ValueError(f"Source path ({source_path}) not found")

        self.logger.info(f"Moving {source_path} to {target_path}...")

        # convert args to str...Path-like objects are supported 3.9+
        out = shutil.move(str(source_path), str(target_path))
        return Path(out)


class Copy(Task):
    """
    Task for copying files or directories within the file system.

    Args:
        - source_path (Union[str, Path], optional): the path to the source directory/file.
        - target_path (Union[str, Path], optional): the path to the target directory/file.
            If copying a directory: the `target_path` must not exists.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        source_path: Union[str, Path] = "",
        target_path: Union[str, Path] = "",
        **kwargs: Any,
    ):
        self.source_path = source_path
        self.target_path = target_path
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "source_path",
        "target_path",
    )
    def run(
        self,
        source_path: Union[str, Path] = "",
        target_path: Union[str, Path] = "",
    ) -> Path:
        """
        Task run method.

        Args:
            - source_path (Union[str, Path], optional): the path to the source directory/file.
            - target_path (Union[str, Path], optional): the path to the target directory/file.
                If copying a directory: the `target_path` must not exists.

        Returns:
            - Path: resulting path of the copied file / directory
        """
        if not source_path:
            raise ValueError("No `source_path` provided.")
        if not target_path:
            raise ValueError("No `target_path` provided.")

        source_path = Path(source_path)
        target_path = Path(target_path)

        if not source_path.exists():
            raise ValueError(f"Source path ({source_path}) not found")

        self.logger.info(f"Copying {source_path} to {target_path}...")

        if source_path.is_file():
            out = shutil.copy(source_path, target_path)
        else:
            out = shutil.copytree(source_path, target_path)
        return Path(out)


class Remove(Task):
    """
    Task for removing files or directories within the file system.

    Args:
        - path (Union[str, Path], optional): file or directory to be removed
            If deleting a directory, the directory must be empty.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        path: Union[str, Path] = "",
        **kwargs: Any,
    ):
        self.path = path
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "path",
    )
    def run(self, path: Union[str, Path] = "") -> None:
        """
        Task run method.

        Args:
            - path (Union[str, Path], optional): file or directory to be removed
        """
        if not path:
            raise ValueError("No `path` provided.")

        path = Path(path)

        if not path.exists():
            raise ValueError(f"Path ({path}) not found")

        self.logger.info(f"Removing {path}...")

        if path.is_file():
            path.unlink()
        else:
            path.rmdir()


class Glob(Task):
    """
    This task returns the content of a given directory matching a given pattern within the file system.
    Args:
        - path (Union[str, Path], optional): directory path
        - pattern (str, optional): glob pattern, defaults to *
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        path: Union[str, Path] = "",
        pattern: str = "*",
        **kwargs: Any,
    ):
        self.path = path
        self.pattern = pattern
        super().__init__(**kwargs)

    @defaults_from_attrs("path", "pattern")
    def run(
        self,
        path: Union[str, Path] = "",
        pattern: str = "*",
    ) -> List[Path]:
        """
        Task run method.

        Args:
            - path (Union[str, Path], optional): directory path to search, defaults to ""
            which targets the current directory
            - pattern (str, optional): glob search pattern, defaults to "*" which lists all files
            in the path
            For more details on pattern options, refer to the
            [pathlib documentation](https://docs.python.org/3/library/pathlib.html#pathlib.Path.glob)

        Returns:
            - list: content of the given path as Path objects
        """
        if not path:
            raise ValueError("No `path` provided.")

        globpath = Path(path).joinpath(pattern)
        self.logger.debug(f"Glob path: {globpath}")

        return list(path.glob(pattern))



"""
Results represent Prefect Task inputs and outputs.  In particular, anytime a
Task runs, its output is encapsulated in a `Result` object.  This object retains
information about what the data is, and how to "handle" it if it needs to be
saved / retrieved at a later time (for example, if this Task requests for its
outputs to be cached or checkpointed).

An instantiated Result object has the following attributes:

- a `value`: the value of a Result represents a single piece of data
- a `safe_value`: this attribute maintains a reference to a `SafeResult` object
    which contains a "safe" representation of the `value`; for example, the
    `value` of a `SafeResult` might be a URI or filename pointing to where the
    raw data lives
- a `serializer`: an object that can serialize Python objects to bytes and
  recover them later

To distinguish between a Task that runs but does not return output from a Task
that has yet to run, Prefect also provides a `NoResult` object representing the
_absence_ of computation / data.  This is in contrast to a `Result` whose value
is `None`.
"""
import copy
import pendulum
import uuid
from typing import Any

from prefect.engine.serializers import PickleSerializer, Serializer
from prefect.utilities import logging


# Subclass of `NotImplementedError` to make it easier to distinguish this error
# in consuming code
class ResultNotImplementedError(NotImplementedError):
    """Indicates a Result feature isn't implemented"""


class Result:
    """
    A representation of the result of a Prefect task; this class contains
    information about the value of a task's result, a result handler specifying
    how to serialize or store this value securely, and a `safe_value` attribute
    which holds information about the current "safe" representation of this
    result.

    Args:
        - value (Any, optional): the value of the result
        - location (Union[str, Callable], optional): Possibly templated location
            to be used for saving the result to the destination. If a callable
            function is provided, it should have signature `callable(**kwargs) ->
            str` and at write time all formatting kwargs will be passed and a fully
            formatted location is expected as the return value.  Can be used for
            string formatting logic that `.format(**kwargs)` doesn't support
        - serializer (Serializer): a serializer that can transform Python
            objects to bytes and recover them. The serializer is used whenever the
            `Result` is writing to or reading from storage. Defaults to
            `PickleSerializer`.
    """

    def __init__(
        self,
        value: Any = None,
        location: str = None,
        serializer: Serializer = None,
    ):
        self.value = value
        if serializer is None:
            serializer = PickleSerializer()
        self.serializer = serializer
        if isinstance(location, (str, type(None))):
            self.location = location
            self._formatter = None
        else:
            self._formatter = location
            self.location = None
        self.logger = logging.get_logger(type(self).__name__)

    def __eq__(self, other: Any) -> bool:
        if type(self) == type(other):
            eq = True
            for attr in self.__dict__:
                if attr.startswith("_"):
                    continue
                eq &= getattr(self, attr, object()) == getattr(other, attr, object())
            return eq
        return False

    def __repr__(self) -> str:
        val = self.value  # type: ignore
        return "<{type}: {val}>".format(type=type(self).__name__, val=repr(val))

    def from_value(self, value: Any) -> "Result":
        """
        Create a new copy of the result object with the provided value.

        Args:
            - value (Any): the value to use

        Returns:
            - Result: a new Result instance with the given value
        """
        new = self.copy()
        new.location = None
        new.value = value
        return new

    def copy(self) -> "Result":
        """
        Return a copy of the current result object.
        """
        return copy.copy(self)

    @property
    def default_location(self) -> str:
        date = pendulum.now("utc").format("Y/M/D")  # type: ignore
        location = f"{date}/{uuid.uuid4()}.prefect_result"
        return location

    def format(self, **kwargs: Any) -> "Result":
        """
        Takes a set of string format key-value pairs and renders the result.location to a final
        location string

        Args:
            - **kwargs (Any): string format arguments for result.location

        Returns:
            - Result: a new result instance with the appropriately formatted location
        """
        new = self.copy()
        if isinstance(new.location, str):
            assert new.location is not None
            new.location = new.location.format(**kwargs)
        elif new._formatter is not None:
            new.location = new._formatter(**kwargs)
        else:
            new.location = new.default_location
        return new

    def exists(self, location: str, **kwargs: Any) -> bool:
        """
        Checks whether the target result exists.

        Args:
            - location (str, optional): Location of the result in the specific result target.
                If provided, will check whether the provided location exists;
                otherwise, will use `self.location`
            - **kwargs (Any): string format arguments for `location`

        Returns:
            - bool: whether or not the target result exists.
        """
        raise ResultNotImplementedError(
            "Not implemented on the base Result class - if you are seeing this error you "
            "might be trying to use features that require choosing a Result subclass; "
            "see https://docs.prefect.io/core/concepts/results.html"
        )

    def read(self, location: str) -> "Result":
        """
        Reads from the target result and returns a corresponding `Result` instance.

        Args:
            - location (str): Location of the result in the specific result target.

        Returns:
            - Any: The value saved to the result.
        """
        raise ResultNotImplementedError(
            "Not implemented on the base Result class - if you are seeing this error you "
            "might be trying to use features that require choosing a Result subclass; "
            "see https://docs.prefect.io/core/concepts/results.html"
        )

    def write(self, value_: Any, **kwargs: Any) -> "Result":
        """
        Serialize and write the result to the target location.

        Args:
            - value_ (Any): the value to write; will then be stored as the `value` attribute
                of the returned `Result` instance
            - **kwargs (optional): if provided, will be used to format the location template
                to determine the location to write to

        Returns:
            - Result: a new result object with the appropriately formatted location destination
        """
        raise ResultNotImplementedError(
            "Not implemented on the base Result class - if you are seeing this error you "
            "might be trying to use features that require choosing a Result subclass; "
            "see https://docs.prefect.io/core/concepts/results.html"
        )


class NoResultType(Result):
    """
    Backwards compatible type so that states created by new versions of Core can be deserialized by
    old versions.
    """

    pass


NoResult = NoResultType()



from typing import List, Callable

import confluent_kafka
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class KafkaBatchConsume(Task):
    """
    Task for consuming a batch of messages from Kafka topics.

    Args:
        - bootstrap_servers (str, optional): comma separated host and port pairs that are the
            addresses of kafka brokers.
        - group_id (str, required): name of the consumer group the consumer will belong to.
            Must be specified either at init or runtime.
        - topics (List[str], required): list of topic names to consume messages from. Must
            be specified either at init or runtime.
        - request_timeout (float, optional): Maximum time to block waiting for message, event or callback
        - auto_offset_reset (str, optional): configurable offset reset policy
        - message_consume_limit (int, optional): max number of messages to consume before closing
            the consumer
        - kafka_configs (dict, optional): a dict of kafka client configuration properties used to
            construct the consumer.
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
            init method
    """

    def __init__(
        self,
        bootstrap_servers: str = None,
        group_id: str = None,
        topics: List[str] = None,
        request_timeout: float = 1.0,
        auto_offset_reset: str = "earliest",
        message_consume_limit: int = None,
        kafka_configs: dict = None,
        **kwargs,
    ):

        self.bootstrap_servers = bootstrap_servers
        self.group_id = group_id
        self.topics = topics
        self.request_timeout = request_timeout
        self.auto_offset_reset = auto_offset_reset
        self.message_consume_limit = message_consume_limit
        self.kafka_configs = kafka_configs or {}
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "bootstrap_servers",
        "group_id",
        "topics",
        "request_timeout",
        "auto_offset_reset",
        "message_consume_limit",
        "kafka_configs",
    )
    def run(
        self,
        bootstrap_servers: str = None,
        group_id: str = None,
        topics: List[str] = None,
        request_timeout: float = 1.0,
        auto_offset_reset: str = "earliest",
        message_consume_limit: int = None,
        kafka_configs: dict = None,
        **kwargs,
    ) -> List[bytes]:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a
        Flow context, or by using `Task.bind`.
        Args:
            - bootstrap_servers (str, required): comma separated host and port pairs that are the
                addresses of kafka brokers
            - group_id (str, required): name of the consumer group the consumer will belong to
            - topics (List[str], required): list of topic names to consume messages from
            - request_timeout (float, optional): Maximum time to block waiting for message, event
                or callback
            - auto_offset_reset (str, optional): configurable offset reset policy
            - message_consume_limit (int, optional): max number of messages to consume before
                closing the consumer
            - kafka_configs (dict, optional): a dict of kafka client configuration properties used
                to construct the consumer.
            - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
                init method
        Returns:
            - List of consumed messages
        """
        consumer = confluent_kafka.Consumer(
            {
                "bootstrap.servers": bootstrap_servers,
                "group.id": group_id,
                "auto.offset.reset": auto_offset_reset,
                **kafka_configs,
            }
        )
        consumer.subscribe(topics)

        messages = []
        message_consume_count = 0
        running = True

        try:
            while running:
                message = consumer.poll(timeout=request_timeout)

                if message is not None:
                    if message.error():
                        if (
                            message.error().code()
                            == confluent_kafka.KafkaError._PARTITION_EOF
                        ):
                            # End of partition event, exit consumer
                            self.logger.warn(
                                f"{message.topic()} [{message.partition()}] "
                                f"reached end at offset {message.offset()}"
                            )
                            running = False
                        elif message.error():
                            raise confluent_kafka.KafkaException(message.error())
                    else:
                        messages.append(message.value())
                        message_consume_count += 1

                        if message_consume_limit:
                            if message_consume_count >= message_consume_limit:
                                break
                else:
                    self.logger.info(
                        f"No messages found for topic {topics}; closing consumer..."
                    )
                    break

        finally:
            consumer.close()

        return messages


class KafkaBatchProduce(Task):
    """
    Task for producing a batch of messages to a Kafka topic.

    Args:
        - bootstrap_servers (str, required): comma separated host and port pairs that are the
            addresses of kafka brokers
        - topic (str, required): name of topic to produce messages to. Must be specified
            either at init or runtime.
        - messages (List[dict], required): list of messages to produce into a topic where
            a single message is a dictionary with a key and a value.
        - flush_threshold (int, optional): threshold of messages produced before flushing
        - callback (Callable, optional): callback assigned to a produce call
        - kafka_configs (dict, optional): a dict of kafka client configuration properties used to
            construct the producer.
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
            init method
    """

    def __init__(
        self,
        bootstrap_servers: str = None,
        topic: str = None,
        messages: List[dict] = None,
        flush_threshold: int = None,
        callback: Callable = None,
        kafka_configs: dict = None,
        **kwargs,
    ):
        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.messages = messages
        self.flush_threshold = flush_threshold
        self.callback = callback
        self.kafka_configs = kafka_configs or {}
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "bootstrap_servers",
        "topic",
        "messages",
        "flush_threshold",
        "callback",
        "kafka_configs",
    )
    def run(
        self,
        bootstrap_servers: str = None,
        topic: str = None,
        messages: List[dict] = None,
        flush_threshold: int = None,
        callback: Callable = None,
        kafka_configs: dict = None,
    ):
        """
        Run method for this Task. Invoked by calling this Task after initialization within a
        Flow context, or by using `Task.bind`.
        Args:
        - bootstrap_servers (str, required): comma separated host and port pairs that are the
            addresses of kafka brokers
        - topic (str, required): name of topics to produce messages to
        - messages (List[dict], required): list of messages to produce into topics where
            a single message is a dictionary with a key and a value.
        - flush_threshold (int, optional): threshold of messages produced before flushing
        - callback (Callable, optional): callback assigned to a produce call
        - kafka_configs (dict, optional): a dict of kafka client configuration properties used to
            construct the producer.
        """

        producer = confluent_kafka.Producer(
            {"bootstrap.servers": bootstrap_servers, **kafka_configs}
        )
        message_produce_count = 0

        for i, message in enumerate(messages):
            if flush_threshold:
                if i % flush_threshold == 0:
                    producer.flush()
                    self.logger.info(
                        f"Producer flushed {flush_threshold} messages to {topic}"
                    )
                    message_produce_count = 0

            key = message.get("key")
            value = message.get("value")

            producer.produce(topic=topic, key=key, value=value, callback=callback)
            message_produce_count += 1

        producer.flush()
        self.logger.info(
            f"Producer flushed {message_produce_count} messages to {topic}"
        )



"""
Utility functions for actually executing flow runs once they've been deployed to an
execution environment

These are not intended to be user-facing.
"""
import copy
import subprocess
import sys
import pendulum
import os
import time
from contextlib import contextmanager
from typing import Any, Type, Optional, cast, Dict

import prefect
from prefect import Flow
from prefect.backend.flow_run import FlowRunView
from prefect.run_configs import RunConfig
from prefect.utilities.logging import get_logger
from prefect.utilities.graphql import with_args
from prefect.configuration import to_environment_variables


logger = get_logger("backend.execution")


def execute_flow_run_in_subprocess(
    flow_run_id: str, run_api_key: str = None, include_local_env: bool = True
) -> FlowRunView:
    """
    Execute a flow run in a subprocess.

    See `execute_flow_run` for more details.

    Intended for executing a flow run without an agent.

    If an interrupt is received during execution, the flow run is marked as failed in
    contrast to an agent-based execution model where the flow run can be resubmitted by
    Lazarus.

    Args:
        - flow_run_id: The flow run to execute
        - run_api_key: The authentication key to provide to the flow run for
            communicating with the Prefect Cloud API. If not set, it will be inferred
            from the current config.
        - include_local_env: If `True`, the currently available environment variables
            will be passed through to the flow run. Defaults to `True` to match
            subprocess behavior.

    Returns:
        FlowRunView: The final flow run object
    """
    flow_run = FlowRunView.from_flow_run_id(flow_run_id)

    # Set up the environment
    env = generate_flow_run_environ(
        flow_run_id=flow_run_id,
        flow_id=flow_run.flow_id,
        run_api_key=run_api_key,
        run_config=flow_run.run_config,
        include_local_env=include_local_env,
    )

    if not flow_run.state.is_scheduled():
        raise RuntimeError(
            f"Flow run is already in state {flow_run.state!r}! Cannot execute flow run "
            "unless it is in a 'Scheduled' state"
        )

    while not flow_run.state.is_finished():

        _wait_for_flow_run_start_time(flow_run_id)

        try:
            # Creating a subprocess allows us to pass environment variables and ensure
            # the flow run loads a correct Prefect config. This CLI command calls out to
            # `execute_flow_run`
            logger.info("Creating subprocess to execute flow run...")
            result = subprocess.run(
                [sys.executable, "-m", "prefect", "execute", "flow-run"], env=env
            )
        except KeyboardInterrupt:
            _fail_flow_run(
                flow_run_id=flow_run_id,
                message="Flow run received an interrupt signal.",
            )
            raise  # Reraise interrupts
        except Exception as exc:
            message = "Flow run encountered unexpected exception during execution"
            _fail_flow_run(
                flow_run_id=flow_run_id,
                message=f"{message}: {exc!r}",
            )
            raise RuntimeError(message) from exc

        logger.debug("Exited flow run subprocess.")

        try:
            result.check_returncode()
        except subprocess.CalledProcessError as exc:
            # Wrap this error to provide a clearer one
            raise RuntimeError("The flow run process failed.") from exc

        # Wait half a second before getting the latest data to reduce the chance of
        # starting another subprocess that exits immediately
        time.sleep(0.5)
        flow_run = flow_run.get_latest()

    return flow_run


def execute_flow_run(
    flow_run_id: str,
    flow: "Flow" = None,
    runner_cls: Type["prefect.engine.flow_runner.FlowRunner"] = None,
    **kwargs: Any,
) -> "FlowRunView":
    """
    TODO: This is not used yet and will be hooked up to the CLI in the near future
          At that point, tests should be written and it should be fully compatible with
          the inline CLI functionality

    The primary entry point for executing a flow run. The flow run will be run
    in-process using the given `runner_cls` which defaults to the `CloudFlowRunner`.

    This function assumes that the flow run's environment variables have been set
    already. See `execute_flow_run_in_subprocess` if you neeed the environment to be
    created.

    Args:
        - flow_run_id: The flow run id to execute; this run id must exist in the database
        - flow: A Flow object can be passed to execute a flow without loading it from
            Storage. If `None`, the flow's Storage metadata will be pulled from the
            API and used to get a functional instance of the Flow and its tasks.
        - runner_cls: An optional `FlowRunner` to override the default `CloudFlowRunner`
        - **kwargs: Additional kwargs will be passed to the `FlowRunner.run` method

    Returns:
        A `FlowRunView` instance with information about the state of the flow run and its
        task runs
    """
    logger.debug(f"Querying for flow run {flow_run_id!r}")

    # Get the `FlowRunner` class type
    # TODO: Respect a config option for this class so it can be overridden by env var,
    #       create a separate config argument for flow runs executed with the backend
    runner_cls = runner_cls or prefect.engine.cloud.flow_runner.CloudFlowRunner

    # Get a `FlowRunView` object
    flow_run = FlowRunView.from_flow_run_id(
        flow_run_id=flow_run_id, load_static_tasks=False
    )

    logger.info(f"Constructing execution environment for flow run {flow_run_id!r}")

    # Populate global secrets
    secrets = prefect.context.get("secrets", {})
    flow_metadata = flow_run.get_flow_metadata()
    if flow_metadata.storage:
        logger.info("Loading secrets...")
        for secret in flow_metadata.storage.secrets:
            with _fail_flow_run_on_exception(
                flow_run_id=flow_run_id,
                message=f"Failed to load flow secret {secret!r}: {{exc}}",
            ):
                secrets[secret] = prefect.tasks.secrets.PrefectSecret(name=secret).run()

    # Load the flow from storage if not explicitly provided
    if not flow:
        logger.info(f"Loading flow from {flow_metadata.storage}...")
        with prefect.context(secrets=secrets, loading_flow=True):
            with _fail_flow_run_on_exception(
                flow_run_id=flow_run_id,
                message="Failed to load flow from storage: {exc}",
            ):
                flow = flow_metadata.storage.get_flow(flow_metadata.name)

    # Update the run context to include secrets with merging
    run_kwargs = copy.deepcopy(kwargs)
    run_kwargs["context"] = run_kwargs.get("context", {})
    run_kwargs["context"]["secrets"] = {
        # User provided secrets will override secrets we pulled from storage and the
        # current context
        **secrets,
        **run_kwargs["context"].get("secrets", {}),
    }
    # Update some default run kwargs with flow settings
    run_kwargs.setdefault("executor", flow.executor)

    # Execute the flow, this call will block until exit
    logger.info(
        f"Beginning execution of flow run {flow_run.name!r} from flow {flow_metadata.name!r} "
        f"with {runner_cls.__name__!r}"
    )
    with prefect.context(flow_run_id=flow_run_id):
        with _fail_flow_run_on_exception(
            flow_run_id=flow_run_id,
            message="Failed to execute flow: {exc}",
        ):
            if flow_metadata.run_config is not None:
                runner_cls(flow=flow).run(**run_kwargs)

            # Support for deprecated `flow.environment` use
            else:
                environment = flow.environment
                environment.setup(flow)
                environment.execute(flow)

    # Get the final state
    flow_run = flow_run.get_latest()
    logger.info(f"Run finished with final state {flow_run.state}")
    return flow_run


def generate_flow_run_environ(
    flow_run_id: str,
    flow_id: str,
    run_config: RunConfig,
    run_api_key: str = None,
    include_local_env: bool = False,
) -> Dict[str, str]:
    """
    Utility to generate the environment variables required for a flow run

    Args:
        - flow_run_id: The id for the flow run that will be executed
        - flow_id: The id for the flow of the flow run that will be executed
        - run_config: The run config for the flow run, contributes environment variables
        - run_api_key: An optional API key to pass to the flow run for authenticating
            with the backend. If not set, it will be pulled from the Client
        - include_local_env: If `True`, the currently available environment variables
            will be passed through to the flow run. Defaults to `False` for security.

    Returns:
        - A dictionary of environment variables
    """
    # TODO: Generalize this and use it for all agents

    # Local environment
    env = cast(Dict[str, Optional[str]], os.environ.copy() if include_local_env else {})

    # Pass through config options that can be overriden by run config
    env.update(
        to_environment_variables(
            prefect.config,
            include={
                "logging.level",
                "logging.format",
                "logging.datefmt",
                "cloud.send_flow_run_logs",
            },
        )
    )

    # Update with run config environment
    if run_config is not None and run_config.env is not None:
        env.update(run_config.env)

    # Update with config options that cannot be overriden by the run config
    env.update(
        to_environment_variables(
            prefect.config,
            include={"backend", "cloud.api", "cloud.tenant_id"},
        )
    )

    # Pass authentication through
    client = prefect.Client()  # Instantiate a client to get the current API key
    env["PREFECT__CLOUD__API_KEY"] = run_api_key or client.api_key or ""
    # Backwards compat for auth tokens
    env["PREFECT__CLOUD__AUTH_TOKEN"] = (
        run_api_key
        or prefect.config.cloud.agent.get("auth_token")
        or prefect.config.cloud.get("auth_token")
    )

    # Add context information for the run
    env.update(
        {
            "PREFECT__CONTEXT__FLOW_RUN_ID": flow_run_id,
            "PREFECT__CONTEXT__FLOW_ID": flow_id,
        }
    )

    # Update hardcoded execution variables
    env.update(
        {
            "PREFECT__ENGINE__FLOW_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudFlowRunner",
            "PREFECT__ENGINE__TASK_RUNNER__DEFAULT_CLASS": "prefect.engine.cloud.CloudTaskRunner",
        }
    )

    # Filter out `None` values
    return {k: v for k, v in env.items() if v is not None}


def _wait_for_flow_run_start_time(flow_run_id: str) -> None:
    """
    Utility that sleeps until the correct flow run start time using
    `_get_flow_run_scheduled_start_time` then `_get_next_task_run_start_time`

    See those two docstrings for details on why this is necessary.
    """
    logger.debug("Checking for flow run scheduled start time...")
    flow_run_start = _get_flow_run_scheduled_start_time(flow_run_id)
    if flow_run_start:
        interval = flow_run_start.diff(abs=False).in_seconds() * -1
        message = f"Flow run scheduled to run {flow_run_start.diff_for_humans()}; "
        if interval > 0:
            logger.info(message + "sleeping before trying to run flow...")
            time.sleep(interval)
        else:  # is scheduled in the past
            logger.debug(message + "moving to next check...")
    else:
        logger.debug("No scheduled time found; moving to next check...")

    logger.debug("Checking for retried task runs...")
    next_task_start = _get_next_task_run_start_time(flow_run_id)
    if next_task_start:
        interval = next_task_start.diff(abs=False).in_seconds() * -1
        message = f"Found task run scheduled {next_task_start.diff_for_humans()}; "
        if interval > 0:
            logger.info(message + "sleeping before trying to run flow...")
            # TODO: In the future we may want to set an upper limit if the backend
            #       may update the state of the flow for an earlier retry
            time.sleep(interval)
        else:  # is scheduled in the past
            logger.info(message + "trying to run flow now...")
    else:
        logger.debug("No scheduled task runs found; trying to run flow now...")


def _get_next_task_run_start_time(flow_run_id: str) -> Optional[pendulum.DateTime]:
    """
    Queries task runs associated with a flow run to get the earliest state start time.
    This time _may_ be in the past.

    Long retries are handled by exiting flow execution leaving the flow run in a
    'Running' state and attaching a start time to the task runs that need to be retried.
    This function checks for a long retry by querying for task runs that have a start
    time set. This allows us to wait until this run time is reached before starting
    flow run execution. If we started execution, the runner would just walk the DAG and
    exit since the task run is not ready to begin yet.

    Args:
        - flow_run_id: The flow run the task runs belong to

    Returns:
        None: If no scheduled task runs are found, otherwise
        pendulum.DateTime: The earliest scheduled task run start time.
    """
    client = prefect.Client()
    result = client.graphql(
        {
            "query": {
                with_args(
                    "task_run",
                    {
                        "where": {
                            "state_start_time": {"_is_null": False},
                            "flow_run_id": {"_eq": flow_run_id},
                            "flow_run": {
                                # Only include flow runs in a 'Running' state to reduce
                                # the scope of the query to retrying flow runs
                                "state": {"_eq": "Running"}
                            },
                        }
                    },
                ): {"state_start_time"}
            }
        }
    )
    task_runs = result.get("data", {}).get("task_run")
    if task_runs is None:
        raise ValueError(f"Unexpected result while querying for task runs: {result}")
    elif not task_runs:
        return None  # No scheduled task runs

    task_run = min(task_runs, key=lambda task_run: task_run.state_start_time)
    next_start_time = task_run.state_start_time
    return cast(pendulum.DateTime, pendulum.parse(next_start_time))


def _get_flow_run_scheduled_start_time(flow_run_id: str) -> Optional[pendulum.DateTime]:
    """
    Queries for the current scheduled start time of a flow

    Flow runs store a `scheduled_start_time` as their originally scheduled time to
    start. 'Scheduled' flow run states also store a `start_time` that supercedes the
    time on the flow run object itself. For example, if a flow run is scheduled for some
    time in the future and a user clicks the 'Start Now' button in the UI, we'll create
    a new 'Scheduled' state with an updated start time. This allows us to preserve
    start time history while making the state the source of truth.

    This function will always return the start time associated with the most recently
    created 'Scheduled' state, if available. If the most recent 'Scheduled' state has a
    null `start_time`, we will fall back to the flow run's `scheduled_start_time`.

    Args:
        - flow_run_id: The flow run of interest

    Returns:
        pendulum.DateTime: The most recent scheduled flow run start time

    Raises:
        - ValueError: On API error
        - ValueError: When zero or more than one flow runs are found

    """
    client = prefect.Client()
    result = client.graphql(
        {
            # We cannot query for states directly and must go through the `flow_run`
            # object
            "query": {
                with_args("flow_run", {"where": {"id": {"_eq": flow_run_id}}}): {
                    with_args("states", {"where": {"state": {"_eq": "Scheduled"}}}): {
                        "start_time",
                        "created",
                    },
                    "scheduled_start_time": True,
                }
            }
        }
    )
    flow_runs = result.get("data", {}).get("flow_run")
    if flow_runs is None:
        raise ValueError(
            f"Unexpected result while querying for flow run states: {result}"
        )
    elif len(flow_runs) > 1:
        raise ValueError(
            f"Found more than one flow run matching id {flow_run_id!r}: {result}"
        )
    elif not flow_runs:
        raise ValueError(f"No flow run exists with id {flow_run_id!r}.")

    # Get the one found flow run
    flow_run = flow_runs[0]

    # Get the most recently created state
    states = sorted(
        flow_run.states, key=lambda state: state.get("created", ""), reverse=True
    )
    state = states[0] if states else None

    # Return the most recently created state start time; default to the flow run
    # scheduled start time in case there are no state times
    start_time = (
        state.start_time
        if state and state.get("start_time")
        else flow_run.scheduled_start_time
    )

    if not start_time:
        return None  # There is no scheduled start time in the states or on the run

    return cast(pendulum.DateTime, pendulum.parse(start_time))


@contextmanager
def _fail_flow_run_on_exception(
    flow_run_id: str,
    message: str = None,
) -> Any:
    """
    A utility context manager to set the state of the given flow run to 'Failed' if
    an exception occurs. A custom message can be provided for more details and will
    be attached to the state and added to the run logs. All errors will be re-raised.



    Args:
        - flow_run_id: The flow run id to update the state of
        - message: The message to include in the state and logs. `{exc}` will be formatted
            with the exception details.

    """
    message = message or "Flow run failed with {exc}"

    try:
        yield
    except Exception as exc:
        message = message.format(exc=exc.__repr__())
        if not FlowRunView.from_flow_run_id(flow_run_id).state.is_finished():
            _fail_flow_run(flow_run_id, message=message)
        logger.error(message, exc_info=True)
        raise


def _fail_flow_run(flow_run_id: str, message: str) -> None:
    """
    Set a flow run to a 'Failed' state and write a a failure message log
    """
    client = prefect.Client()
    client.set_flow_run_state(
        flow_run_id=flow_run_id, state=prefect.engine.state.Failed(message)
    )
    client.write_run_logs(
        [
            dict(
                flow_run_id=flow_run_id,  # type: ignore
                name="prefect.backend.execution",
                message=message,
                level="ERROR",
            )
        ]
    )



from typing import TYPE_CHECKING, List, Callable

if TYPE_CHECKING:
    from click import Option


def add_options(options: List["Option"]) -> Callable:
    """A decorator for adding a list of options to a click command"""

    def decorator(func: Callable) -> Callable:
        for opt in reversed(options):
            func = opt(func)  # type: ignore
        return func

    return decorator



import os
import ssl
import smtplib
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Any, cast, List

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


class EmailTask(Task):
    """
    Task for sending email from an authenticated email service over SMTP. For this task to
    function properly you must have the `"EMAIL_USERNAME"` and `"EMAIL_PASSWORD"`
    Prefect Secrets set.  It is recommended you use a [Google App
    Password](https://support.google.com/accounts/answer/185833) if you use Gmail.  The default
    SMTP server is set to the Gmail SMTP server on port 465 (SMTP-over-SSL). Sending messages
    containing HTML code is supported - the default MIME type is set to the text/html.

    You can also use `smtp_type="INSECURE"` and `smtp_port=25` to use an insecure, internal SMTP server.
    The `"EMAIL_USERNAME"` and `"EMAIL_PASSWORD"` secrets are not required in this case.

    Args:
        - subject (str, optional): the subject of the email; can also be provided at runtime
        - msg (str, optional): the contents of the email, added as html; can be used in
            combination of msg_plain; can also be provided at runtime
        - email_to (str, optional): the destination email address to send the message to; can also
            be provided at runtime
        - email_from (str, optional): the email address to send from; defaults to
            notifications@prefect.io
        - smtp_server (str, optional): the hostname of the SMTP server; defaults to smtp.gmail.com
        - smtp_port (int, optional): the port number of the SMTP server; defaults to 465
        - smtp_type (str, optional): either SSL, STARTTLS, or INSECURE; defaults to SSL
        - msg_plain (str, optional): the contents of the email, added as plain text can be used in
            combination of msg; can also be provided at runtime
        - email_to_cc (str, optional): additional email address to send the message to as cc;
            can also be provided at runtime
        - email_to_bcc (str, optional): additional email address to send the message to as bcc;
            can also be provided at runtime
        - attachments (List[str], optional): names of files that should be sent as attachment; can
            also be provided at runtime
        - **kwargs (Any, optional): additional keyword arguments to pass to the base Task
            initialization
    """

    def __init__(
        self,
        subject: str = None,
        msg: str = None,
        email_to: str = None,
        email_from: str = "notifications@prefect.io",
        smtp_server: str = "smtp.gmail.com",
        smtp_port: int = 465,
        smtp_type: str = "SSL",
        msg_plain: str = None,
        email_to_cc: str = None,
        email_to_bcc: str = None,
        attachments: List[str] = None,
        **kwargs: Any,
    ):
        self.subject = subject
        self.msg = msg
        self.email_to = email_to
        self.email_from = email_from
        self.smtp_server = smtp_server
        self.smtp_port = smtp_port
        self.smtp_type = smtp_type
        self.msg_plain = msg_plain
        self.email_to_cc = email_to_cc
        self.email_to_bcc = email_to_bcc
        self.attachments = attachments or []
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "subject",
        "msg",
        "email_to",
        "email_from",
        "smtp_server",
        "smtp_port",
        "smtp_type",
        "msg_plain",
        "email_to_cc",
        "email_to_bcc",
        "attachments",
    )
    def run(
        self,
        subject: str = None,
        msg: str = None,
        email_to: str = None,
        email_from: str = None,
        smtp_server: str = None,
        smtp_port: int = None,
        smtp_type: str = None,
        msg_plain: str = None,
        email_to_cc: str = None,
        email_to_bcc: str = None,
        attachments: List[str] = None,
    ) -> None:
        """
        Run method which sends an email.

        Args:
            - subject (str, optional): the subject of the email; defaults to the one provided
                at initialization
            - msg (str, optional): the contents of the email; defaults to the one provided
                at initialization
            - email_to (str, optional): the destination email address to send the message to;
                defaults to the one provided at initialization
            - email_from (str, optional): the email address to send from; defaults to the one
                provided at initialization
            - smtp_server (str, optional): the hostname of the SMTP server; defaults to the one
                provided at initialization
            - smtp_port (int, optional): the port number of the SMTP server; defaults to the one
                provided at initialization
            - smtp_type (str, optional): either SSL, STARTTLS, or INSECURE; defaults to the one provided
                at initialization
            - msg_plain (str, optional): the contents of the email, added as plain text can be used in
                combination of msg; defaults to the one provided at initialization
            - email_to_cc (str, optional): additional email address to send the message to as cc;
                defaults to the one provided at initialization
            - email_to_bcc (str, optional): additional email address to send the message to as bcc;
                defaults to the one provided at initialization
            - attachments (List[str], optional): names of files that should be sent as attachment;
                defaults to the one provided at initialization

        Returns:
            - None
        """

        if smtp_type != "INSECURE":
            username = cast(str, Secret("EMAIL_USERNAME").get())
            password = cast(str, Secret("EMAIL_PASSWORD").get())

        message = MIMEMultipart()
        message["Subject"] = subject
        message["From"] = email_from
        message["To"] = email_to
        if email_to_cc:
            message["Cc"] = email_to_cc
        if email_to_bcc:
            message["Bcc"] = email_to_bcc

        # First add the message in plain text, then the HTML version. Email clients try to render
        # the last part first
        if msg_plain:
            message.attach(MIMEText(msg_plain, "plain"))
        if msg:
            message.attach(MIMEText(msg, "html"))

        for filepath in attachments:
            with open(filepath, "rb") as attachment:
                part = MIMEBase("application", "octet-stream")
                part.set_payload(attachment.read())

            encoders.encode_base64(part)
            filename = os.path.basename(filepath)
            part.add_header(
                "Content-Disposition",
                f"attachment; filename= {filename}",
            )
            message.attach(part)

        if smtp_type == "INSECURE":
            server = smtplib.SMTP(smtp_server, smtp_port)
        else:
            context = ssl.create_default_context()
            if smtp_type == "SSL":
                server = smtplib.SMTP_SSL(smtp_server, smtp_port, context=context)
            elif smtp_type == "STARTTLS":
                server = smtplib.SMTP(smtp_server, smtp_port)
                server.starttls(context=context)
            else:
                raise ValueError(f"{smtp_type} is an unsupported value for smtp_type.")

            server.login(username, password)

        try:
            server.send_message(message)
        finally:
            server.quit()



import time

import click

import prefect
from prefect import config
from prefect.client import Client
from prefect.utilities.logging import get_logger


@click.group(hidden=True)
def heartbeat():
    """
    Send heartbeats back to the Prefect API.

    \b
    Usage:
        $ prefect heartbeat [task-run/flow-run] -i ID

    \b
    Arguments:
        task-run   Send heartbeat for a given task run ID
        flow-run   Send heartbeat for a given flow run ID
    """


@heartbeat.command(hidden=True)
@click.option(
    "--id",
    "-i",
    required=True,
    help="The ID of the flow run to heartbeat.",
    hidden=True,
)
@click.option(
    "--num",
    "-n",
    type=int,
    help="The number of times to heartbeat; if not provided, will heartbeat for forever.",
    hidden=True,
)
def flow_run(id, num):
    """
    Send heartbeats back to the Prefect API for a given flow run ID.

    \b
    Options:
        --id, -i                TEXT        The id of a flow run to send heartbeats for [required]
        --num, -n               TEXT        The number of times to send a heartbeat [required]

    \b
    If num is not provided, the heartbeat will be sent indefinitely.
    """

    client = Client()
    logger = get_logger("subprocess_heartbeat")
    iter_count = 0

    # Ensure that logs are sent to the backend since this is typically called without
    # console stdout/err
    with prefect.context({"flow_run_id": id, "running_with_backend": True}):

        try:  # Log signal-like exceptions that cannot be ignored

            while iter_count < (num or 1):

                try:  # Ignore (but log) client exceptions
                    client.update_flow_run_heartbeat(id)
                except Exception as exc:
                    logger.error(
                        f"Failed to send heartbeat with exception: {exc!r}",
                        exc_info=True,
                    )

                if num:
                    iter_count += 1
                time.sleep(config.cloud.heartbeat_interval)

        except BaseException as exc:
            logger.error(
                f"Heartbeat process encountered terminal exception: {exc!r}",
                exc_info=True,
            )
            raise



from typing import Any

from dateutil import rrule
from marshmallow import fields, post_dump, post_load
import pendulum
import prefect
from prefect.utilities.serialization import (
    DateTimeTZ,
    JSONCompatible,
    ObjectSchema,
    OneOfSchema,
    StatefulFunctionReference,
)

FILTERS = [
    prefect.schedules.filters.on_datetime,
    prefect.schedules.filters.on_date,
    prefect.schedules.filters.at_time,
    prefect.schedules.filters.between_datetimes,
    prefect.schedules.filters.between_dates,
    prefect.schedules.filters.between_times,
    prefect.schedules.filters.is_weekday,
    prefect.schedules.filters.is_weekend,
    prefect.schedules.filters.is_month_end,
    prefect.schedules.filters.is_day_of_week,
    prefect.schedules.filters.is_month_start,
]

ADJUSTMENTS = [
    prefect.schedules.adjustments.add,
    prefect.schedules.adjustments.next_weekday,
]


class IntervalClockSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.clocks.IntervalClock

    start_date = DateTimeTZ(allow_none=True)
    end_date = DateTimeTZ(allow_none=True)
    interval = fields.TimeDelta(precision="microseconds", required=True)
    parameter_defaults = fields.Dict(
        keys=fields.Str(), values=JSONCompatible(), allow_none=True
    )
    labels = fields.List(fields.Str(), allow_none=True)

    @post_dump
    def _interval_validation(self, data: dict, **kwargs: Any) -> dict:
        """
        Ensures interval is at least one minute in length
        """
        if data["interval"] / 1e6 < 60:
            raise ValueError(
                "Interval can not be less than one minute when deploying to Prefect Cloud."
            )
        return data

    @post_load
    def create_object(self, data: dict, **kwargs: Any):
        if data["interval"].total_seconds() < 60:
            raise ValueError(
                "Interval can not be less than one minute when deploying to Prefect Cloud."
            )
        base_obj = super().create_object(data)
        return base_obj


class CronClockSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.clocks.CronClock

    start_date = DateTimeTZ(allow_none=True)
    end_date = DateTimeTZ(allow_none=True)
    cron = fields.String(required=True)
    parameter_defaults = fields.Dict(
        keys=fields.Str(), values=JSONCompatible(), allow_none=True
    )
    labels = fields.List(fields.Str(), allow_none=True)
    day_or = fields.Boolean(allow_none=True)


class DatesClockSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.clocks.DatesClock

    dates = fields.List(DateTimeTZ(), required=True)
    parameter_defaults = fields.Dict(
        keys=fields.Str(), values=JSONCompatible(), allow_none=True
    )
    labels = fields.List(fields.Str(), allow_none=True)


class RRuleSchema(ObjectSchema):
    class Meta:
        object_class = rrule.rrule

    rr = fields.Method("dump_rrule_str", "load_rrule_str")
    # RRule str serialization does not record the timezone for some weird reason. So, in order to
    # get it back we also serialize the start/until params which may have TZ, and replace them
    _dtstart = fields.DateTime(
        allow_none=True, load_default=pendulum.datetime(2018, 6, 20)
    )
    _until = fields.DateTime(allow_none=True)

    def dump_rrule_str(self, obj):
        return str(obj)

    def load_rrule_str(self, value):
        return rrule.rrulestr(value)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Any:
        rr = data["rr"]
        params = {"dtstart": data["_dtstart"]}
        if "_until" in data:
            params["until"] = data["_until"]
        rr = rr.replace(**params)
        return rr


class RRuleSetSchema(ObjectSchema):
    class Meta:
        object_class = rrule.rruleset

    _rrule = fields.List(fields.Nested(RRuleSchema))
    _rdate = fields.List(fields.DateTime)
    _exrule = fields.List(fields.Nested(RRuleSchema))
    _exdate = fields.List(fields.DateTime)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Any:
        rrs = rrule.rruleset()
        for rr in data.get("_rrule", []):
            rrs.rrule(rr)
        for dt in data.get("_rdate", []):
            rrs.rdate(dt)
        for exrr in data.get("_exrule", []):
            rrs.exrule(exrr)
        for exdt in data.get("_exdate", []):
            rrs.exdate(exdt)
        return rrs


class RRuleBaseSchema(OneOfSchema):
    type_schemas = {"rrule": RRuleSchema, "rruleset": RRuleSetSchema}


class RRuleClockSchema(ObjectSchema):
    class Meta:
        object_class = prefect.schedules.clocks.RRuleClock

    rrule_obj = fields.Nested(RRuleBaseSchema)
    start_date = DateTimeTZ(allow_none=True)
    end_date = DateTimeTZ(allow_none=True)
    parameter_defaults = fields.Dict(
        keys=fields.Str(), values=JSONCompatible(), allow_none=True
    )
    labels = fields.List(fields.Str(), allow_none=True)


class ClockSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "IntervalClock": IntervalClockSchema,
        "CronClock": CronClockSchema,
        "DatesClock": DatesClockSchema,
        "RRuleClock": RRuleClockSchema,
    }


class ScheduleSchema(ObjectSchema):
    """
    The serialization schema for Schedule types
    """

    class Meta:
        object_class = prefect.schedules.Schedule

    clocks = fields.Nested(ClockSchema, required=True, many=True)
    filters = fields.List(
        StatefulFunctionReference(
            valid_functions=FILTERS, reject_invalid=True, allow_none=True
        )
    )
    or_filters = fields.List(
        StatefulFunctionReference(
            valid_functions=FILTERS, reject_invalid=True, allow_none=True
        )
    )
    not_filters = fields.List(
        StatefulFunctionReference(
            valid_functions=FILTERS, reject_invalid=True, allow_none=True
        )
    )
    adjustments = fields.List(
        StatefulFunctionReference(
            valid_functions=ADJUSTMENTS, reject_invalid=True, allow_none=True
        )
    )



import click

from prefect.client import Client


@click.group(hidden=True)
def delete():
    """
    Delete commands that refer to mutations of Prefect API metadata.

    \b
    Usage:
        $ prefect delete [OBJECT]

    \b
    Arguments:
        project    Delete projects

    \b
    Examples:
        $ prefect delete project "Goodbye, World!"
        Goodbye, World! deleted

    """


@delete.command(hidden=True)
@click.argument("name", required=True)
def project(name):
    """
    Delete projects with the Prefect API that organize flows.

    \b
    Arguments:
        name                TEXT    The name of a project to delete     [required]

    """
    try:
        Client().delete_project(project_name=name)
    except ValueError as exc:
        click.echo(f"{type(exc).__name__}: {exc}")
        click.secho("Error deleting project", fg="red")
        return

    click.secho("{} deleted".format(name), fg="green")



import json
import os
from typing import Dict, Any

import prefect
import platform


def system_information() -> dict:
    """
    Get system information

    Returns:
        - dict: a dictionary containing some system information
    """
    return dict(
        system_information=dict(
            platform=platform.platform(),
            python_version=platform.python_version(),
            prefect_version=prefect.__version__,
            prefect_backend=prefect.config.backend,
        )
    )


def config_overrides(include_secret_names: bool = False) -> dict:
    """
    Get user configuration keys that differ from the default configuration. Will only
    return an indication if a key is set and differs from the defaults, values are
    *not* returned.

    Args:
        - include_secret_names (bool, optional): toggle inclusion of secret config keys
            in the output. Note that secret values are never returned, only their names
            when this is `True`. Defaults to `False`.

    Returns:
        - dict: a dictionary containing names of user configuration overrides
    """
    # Replace values with boolean signalling variable presence
    def _replace_values(data: dict) -> Dict[Any, Any]:
        if isinstance(data, dict):
            return {
                k: _replace_values(v)
                if k != "secrets" or include_secret_names
                else False
                for k, v in data.items()
            }
        return True

    # Load the default config to compare values
    default_config = dict()
    default_config_path = prefect.configuration.DEFAULT_CONFIG
    if default_config_path and os.path.isfile(default_config_path):
        default_config = prefect.configuration.load_toml(default_config_path)

    user_config = dict()  # type: ignore
    user_config_path = prefect.configuration.USER_CONFIG
    if user_config_path and os.path.isfile(
        str(prefect.configuration.interpolate_env_vars(user_config_path))
    ):
        user_config = prefect.configuration.load_toml(user_config_path)

    # Create some shorter names for fully specified imports avoiding circular
    # dependencies in the utilities
    dict_to_flatdict = prefect.utilities.collections.dict_to_flatdict
    flatdict_to_dict = prefect.utilities.collections.flatdict_to_dict

    # Drop keys from `user_config` that have values identical to `default_config`
    # converting to flat dictionaries for ease of comparison
    user_config = dict_to_flatdict(user_config)
    default_config = dict_to_flatdict(default_config)

    # Collect keys to drop in a list so we aren't dropping keys during iteration
    keys_to_drop = [
        key
        for key, val in user_config.items()
        if key in default_config and val == default_config[key]
    ]

    for key in keys_to_drop:
        user_config.pop(key)

    # Restore to a nested dictionary then replace values with bools
    user_config = flatdict_to_dict(user_config)
    user_config = _replace_values(user_config)

    return dict(config_overrides=user_config)


def environment_variables() -> dict:
    """
    Get `PREFECT__` specific environment variables

    Returns:
        - dict: a dictionary containing names of set Prefect environment variables
    """
    env_vars = list()
    for env_var, _ in os.environ.items():
        if env_var.startswith(prefect.configuration.ENV_VAR_PREFIX + "__"):
            env_vars.append(env_var)

    return dict(env_vars=env_vars)


def flow_information(flow: "prefect.Flow") -> dict:
    """
    Get flow information

    Args:
        - flow ("prefect.Flow"): the flow whose attributes to retrieve

    Returns:
        - dict: a dictionary of informative flow attributes
    """

    def _replace_values(data: dict) -> Dict[Any, Any]:
        if isinstance(data, dict):
            return {k: _replace_values(v) if v else False for k, v in data.items()}

        return True

    # Check presence of environment attributes
    if flow.environment:
        environment = {
            "type": type(flow.environment).__name__,  # type: ignore
            **_replace_values(flow.environment.__dict__),  # type: ignore
        }
    else:
        environment = False  # type: ignore

    # Check presence of run_config attributes
    if flow.run_config:
        run_config = {
            "type": type(flow.run_config).__name__,
            **_replace_values(flow.run_config.__dict__),
        }
    else:
        run_config = False  # type: ignore

    # Check presence of storage attributes
    if flow.storage:
        storage = {
            "type": type(flow.storage).__name__,
            **_replace_values(flow.storage.__dict__),
        }
    else:
        storage = False  # type: ignore

    # Check presence of a result handler
    if flow.result:
        result = {"type": type(flow.result).__name__}
    else:
        result = False  # type: ignore

    # Check presence of a schedule
    if flow.schedule:
        schedule = {"type": type(flow.schedule).__name__, **flow.schedule.__dict__}
    else:
        schedule = False  # type: ignore

    return dict(
        flow_information=dict(
            environment=environment,
            run_config=run_config,
            storage=storage,
            result=result,
            schedule=schedule,
            task_count=len(flow.tasks),
        )
    )


def diagnostic_info(
    flow: "prefect.Flow" = None, include_secret_names: bool = False
) -> str:
    """
    Get full diagnostic information

    Args:
        - flow ("prefect.Flow", optional): the flow whose attributes to retrieve. If no
            flow is provided then flow information will not be present in output.
        - include_secret_names (bool, optional): toggle output of Secret names, defaults to False.
            Note: Secret values are never returned, only their names.

    Returns:
        - str: a string representation of the full diagnostic information
    """
    aggregate_info = dict()

    aggregate_info.update(system_information())
    aggregate_info.update(config_overrides(include_secret_names))
    aggregate_info.update(environment_variables())

    if flow:
        aggregate_info.update(flow_information(flow))

    return json.dumps(aggregate_info, sort_keys=True, indent=2)



from dataclasses import dataclass
from typing import List, Any

from prefect import Client
from prefect.utilities.graphql import with_args


@dataclass(frozen=True)
class TenantView:
    """
    A view of tenant data stored in the Prefect API.

    This object is designed to be an immutable view of the data stored in the Prefect
    backend API at the time it is created

    EXPERIMENTAL: This interface is experimental and subject to change

    Args:
        - tenant_id: The uuid of the tenant
        - name: The name of the tenant
        - slug: A machine compatible unique identifier for the tenant
    """

    tenant_id: str
    name: str
    slug: str

    @classmethod
    def _from_tenant_data(cls, tenant_data: dict) -> "TenantView":
        tenant_data = tenant_data.copy()
        tenant_id = tenant_data.pop("id")
        return cls(tenant_id=tenant_id, **tenant_data)

    @classmethod
    def _query_for_tenant(cls, where: dict, **kwargs: Any) -> dict:
        """
        Query for tenant data using `_query_for_tenants` but throw an exception if
        more than one matching tenant is found

        Args:
            - where: The `where` clause to use
            - **kwargs: Additional kwargs are passed to `_query_for_tenants`

        Returns:
            A dict of tenant data
        """
        tenants = cls._query_for_tenants(where=where, **kwargs)

        if len(tenants) > 1:
            raise ValueError(
                f"Found multiple ({len(tenants)}) tenants while querying for tenants "
                f"where {where}: {tenants}"
            )

        if not tenants:
            return {}

        tenant = tenants[0]
        return tenant

    @staticmethod
    def _query_for_tenants(
        where: dict,
        order_by: dict = None,
        error_on_empty: bool = True,
    ) -> List[dict]:
        """
        Query for tenant data necessary to initialize `TenantView` instances with
        `TenantView._from_tenant_data`.

        Args:
            - where (required): The Hasura `where` clause to filter by
            - order_by (optional): An optional Hasura `order_by` clause to order
                 results by
            - error_on_empty (optional): If `True` and no tenants are found, a
                `ValueError` will be raised

        Returns:
            A list of dicts of tenant information
        """
        client = Client()

        query_args = {"where": where}
        if order_by is not None:
            query_args["order_by"] = order_by

        tenant_query = {
            "query": {
                with_args("tenant", query_args): {
                    "id",
                    "slug",
                    "name",
                }
            }
        }

        result = client.graphql(tenant_query)
        tenants = result.get("data", {}).get("tenant", None)

        if tenants is None:
            raise ValueError(
                f"Received bad result while querying for tenants where {where}: "
                f"{result}"
            )

        if not tenants:  # Empty list
            if error_on_empty:
                raise ValueError(
                    f"No results found while querying for tenants where {where!r}"
                )
            return []

        # Return a list
        return tenants

    @classmethod
    def from_tenant_id(cls, tenant_id: str) -> "TenantView":
        """
        Get an instance of this class filled with information by querying for the given
        tenant id

        Args:
            - tenant_id: the tenant to lookup

        Returns:
            A populated `TenantView` instance
        """
        return cls._from_tenant_data(
            cls._query_for_tenant(where={"id": {"_eq": tenant_id}})
        )

    @classmethod
    def from_current_tenant(cls) -> "TenantView":
        """
        Get an instance of this class filled with information by querying for the
        tenant id set in the Prefect Client

        Returns:
            A populated `TenantView` instance
        """
        client = Client()
        return cls.from_tenant_id(client.tenant_id)



"""
This module contains a task for interacting with dbt via the shell.
"""

try:
    from prefect.tasks.dbt.dbt import DbtShellTask, DbtCloudRunJob
except ImportError as err:
    raise ImportError(
        "Using `prefect.tasks.dbt` requires dbt to be installed."
    ) from err



from pathlib import Path
import snowflake.connector as sf
from snowflake.connector.cursor import SnowflakeCursor

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class SnowflakeQuery(Task):
    """
    Task for executing a query against a Snowflake database.

    Args:
        - account (str): snowflake account name, see snowflake connector
             package documentation for details
        - user (str): user name used to authenticate
        - password (str, optional): password used to authenticate.
            password or private_key must be present
        - private_key (bytes, optional): pem to authenticate.
            password or private_key must be present
        - database (str, optional): name of the default database to use
        - schema (int, optional): name of the default schema to use
        - role (str, optional): name of the default role to use
        - warehouse (str, optional): name of the default warehouse to use
        - query (str, optional): query to execute against database
        - data (tuple, optional): values to use in query, must be specified using placeholder
            in query string
        - autocommit (bool, optional): set to True to autocommit, defaults to None, which
            takes snowflake AUTOCOMMIT parameter
        - cursor_type (SnowflakeCursor, optional): specify the type of database
            cursor to use for the query, defaults to SnowflakeCursor
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        account: str = None,
        user: str = None,
        password: str = None,
        private_key: bytes = None,
        database: str = None,
        schema: str = None,
        role: str = None,
        warehouse: str = None,
        query: str = None,
        data: tuple = None,
        autocommit: bool = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
        **kwargs
    ):
        self.account = account
        self.user = user
        self.password = password
        self.private_key = private_key
        self.database = database
        self.schema = schema
        self.role = role
        self.warehouse = warehouse
        self.query = query
        self.data = data
        self.autocommit = autocommit
        self.cursor_type = cursor_type
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "account",
        "user",
        "password",
        "private_key",
        "database",
        "schema",
        "role",
        "warehouse",
        "query",
        "data",
        "autocommit",
        "cursor_type",
    )
    def run(
        self,
        account: str = None,
        user: str = None,
        password: str = None,
        private_key: bytes = None,
        database: str = None,
        schema: str = None,
        role: str = None,
        warehouse: str = None,
        query: str = None,
        data: tuple = None,
        autocommit: bool = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
    ):
        """
        Task run method. Executes a query against snowflake database.

        Args:
            - account (str, optional): snowflake account name, see snowflake connector
                package documentation for details
            - user (str, optional): user name used to authenticate
            - password (str, optional): password used to authenticate.
                password or private_lkey must be present
            - private_key (bytes, optional): pem to authenticate.
                password or private_key must be present
            - database (str, optional): name of the default database to use
            - schema (int, optional): name of the default schema to use
            - role (str, optional): name of the default role to use
            - warehouse (str, optional): name of the default warehouse to use
            - query (str, optional): query to execute against database
            - data (tuple, optional): values to use in query, must be specified using placeholder
                in query string
            - autocommit (bool, optional): set to True to autocommit, defaults to None, which
                takes snowflake AUTOCOMMIT parameter
            - cursor_type (SnowflakeCursor, optional): specify the type of database
                cursor to use for the query, defaults to SnowflakeCursor

        Returns:
            - List[List]: output of cursor.fetchall()

        Raises:
            - ValueError: if a required parameter is not supplied
            - DatabaseError: if exception occurs when executing the query
        """
        if not account:
            raise ValueError("An account must be provided")
        if not user:
            raise ValueError("A user must be provided")
        if not query:
            raise ValueError("A query string must be provided")

        # build the connection parameter dictionary
        # we will remove `None` values next
        connect_params = {
            "account": account,
            "user": user,
            "password": password,
            "private_key": private_key,
            "database": database,
            "schema": schema,
            "role": role,
            "warehouse": warehouse,
            "autocommit": autocommit,
        }

        # filter out unset values
        connect_params = {
            param: value
            for (param, value) in connect_params.items()
            if value is not None
        }

        # connect to database, open cursor
        conn = sf.connect(**connect_params)
        # try to execute query
        # context manager automatically rolls back failed transactions
        try:
            with conn:
                with conn.cursor(cursor_type) as cursor:
                    executed = cursor.execute(query, params=data).fetchall()
            conn.close()
            return executed

        # pass through error, and ensure connection is closed
        except Exception as error:
            conn.close()
            raise error


class SnowflakeQueriesFromFile(Task):
    """
    Task for executing queries loaded from a file against a Snowflake database.
    Return a list containings the results of the queries.

    Note that using execute_string() is vulnerable to SQL injection.

    Args:
        - account (str, optional): snowflake account name, see snowflake connector
             package documentation for details
        - user (str, optional): user name used to authenticate
        - password (str, optional): password used to authenticate.
            password or private_lkey must be present
        - private_key (bytes, optional): pem to authenticate.
            password or private_key must be present
        - database (str, optional): name of the default database to use
        - schema (int, optional): name of the default schema to use
        - role (str, optional): name of the default role to use
        - warehouse (str, optional): name of the default warehouse to use
        - file_path (str, optional): file path to load query from
        - autocommit (bool, optional): set to True to autocommit, defaults to None, which
            takes snowflake AUTOCOMMIT parameter
        - cursor_type (SnowflakeCursor, optional): specify the type of database
            cursor to use for the query, defaults to SnowflakeCursor
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        account: str = None,
        user: str = None,
        password: str = None,
        private_key: bytes = None,
        database: str = None,
        schema: str = None,
        role: str = None,
        warehouse: str = None,
        file_path: str = None,
        autocommit: bool = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
        **kwargs
    ):
        self.account = account
        self.user = user
        self.password = password
        self.private_key = private_key
        self.database = database
        self.schema = schema
        self.role = role
        self.warehouse = warehouse
        self.file_path = file_path
        self.autocommit = autocommit
        self.cursor_type = cursor_type
        super().__init__(**kwargs)

    @defaults_from_attrs(
        "account",
        "user",
        "password",
        "private_key",
        "database",
        "schema",
        "role",
        "warehouse",
        "file_path",
        "autocommit",
        "cursor_type",
    )
    def run(
        self,
        account: str,
        user: str,
        password: str = None,
        private_key: bytes = None,
        database: str = None,
        schema: str = None,
        role: str = None,
        warehouse: str = None,
        file_path: str = None,
        autocommit: bool = None,
        cursor_type: SnowflakeCursor = SnowflakeCursor,
    ):
        """
        Task run method. Executes a query against snowflake database.

        Args:
            - account (str): snowflake account name, see snowflake connector
                package documentation for details
            - user (str): user name used to authenticate
            - password (str, optional): password used to authenticate.
                password or private_lkey must be present
            - private_key (bytes, optional): pem to authenticate.
                password or private_key must be present
            - database (str, optional): name of the default database to use
            - schema (int, optional): name of the default schema to use
            - role (str, optional): name of the default role to use
            - warehouse (str, optional): name of the default warehouse to use
            - file_path (str, optional): file path to load query from
            - autocommit (bool, optional): set to True to autocommit, defaults to None, which
                takes snowflake AUTOCOMMIT parameter
            - cursor_type (SnowflakeCursor, optional): specify the type of database
                cursor to use for the query, defaults to SnowflakeCursor

        Returns:
            - List[List]: containing the results of the different queries executed

        Raises:
            - ValueError: if query parameter is None or a blank string
            - DatabaseError: if exception occurs when executing the query
            - FileNotFoundError: if File does not exist
        """
        if account is None:
            raise ValueError("An account must be provided")
        if user is None:
            raise ValueError("A user must be provided")
        if file_path is None:
            raise ValueError("A file path must be provided")

        # build the connection parameter dictionary
        # we will remove `None` values next
        connect_params = {
            "account": account,
            "user": user,
            "password": password,
            "private_key": private_key,
            "database": database,
            "schema": schema,
            "role": role,
            "warehouse": warehouse,
            "autocommit": autocommit,
        }

        # filter out unset values
        connect_params = {
            param: value
            for (param, value) in connect_params.items()
            if value is not None
        }

        # connect to database, open cursor
        conn = sf.connect(**connect_params)

        # try to execute query
        # context manager automatically rolls back failed transactions
        try:
            # load query from file
            query = Path(file_path).read_text()

            with conn:
                result = []
                cursor_list = conn.execute_string(query, cursor_class=cursor_type)

                for cursor in cursor_list:
                    result.append(cursor.fetchall())
                    # return fetch for each cursor
            return result

        # ensure connection is closed
        finally:
            conn.close()



import sys
import click

from prefect import config
from prefect.backend import kv_store
from prefect.backend.kv_store import NON_CLOUD_BACKEND_ERROR_MESSAGE
from prefect.cli.build_register import (
    handle_terminal_error,
    TerminalError,
    log_exception,
)


@click.group()
def kv():
    """
    Interact with Prefect Cloud KV Store

    \b
    Usage:
        $ prefect kv [COMMAND]
    """
    if config.backend != "cloud":
        click.secho(NON_CLOUD_BACKEND_ERROR_MESSAGE, fg="red")
        sys.exit(1)


@kv.command(name="set")
@click.argument("key")
@click.argument("value")
@handle_terminal_error
def set_command(key, value):
    """
    Set a key value pair, overriding existing values if key exists

    \b
    Arguments:
        key         TEXT    Key to set
        value       TEXT    Value associated with key to set
    """
    try:
        kv_store.set_key_value(key=key, value=value)
        click.secho("Key value pair set successfully", fg="green")
    except Exception as exc:
        log_exception(exc)
        raise TerminalError("An error occurred setting the key value pair")


@kv.command(name="get")
@click.argument("key")
@handle_terminal_error
def get_command(key):
    """
    Get the value of a key

    \b
    Arguments:
        key         TEXT    Key to get
    """
    try:
        result = kv_store.get_key_value(key=key)
        click.secho(f"Key {key!r} has value {result!r}", fg="green")
    except Exception as exc:
        log_exception(exc)
        raise TerminalError(f"Error retrieving value for key {key!r}")


@kv.command(name="delete")
@click.argument("key")
@handle_terminal_error
def delete_command(key):
    """
    Delete a key value pair

    \b
    Arguments:
        key         TEXT    Key to delete
    """
    try:
        kv_store.delete_key(key=key)
        click.secho(f"Key {key!r} has been deleted", fg="green")
    except Exception as exc:
        log_exception(exc)
        raise TerminalError("An error occurred deleting the key")


@kv.command(name="list")
@handle_terminal_error
def list_command():
    """
    List all key value pairs
    """
    try:
        result = kv_store.list_keys()
        if result:
            click.secho("\n".join(result), fg="green")
        else:
            click.secho("No keys found", fg="yellow")
    except Exception as exc:
        log_exception(exc)
        raise TerminalError("An error occurred when listing keys")



from prefect import Task
from typing import Any
from prefect.utilities.tasks import defaults_from_attrs

try:
    import asana
except ImportError:
    pass


class OpenAsanaToDo(Task):
    """
    Task for opening / creating new Asana tasks using the Asana REST API.

    Args:
        - project (str; , required): The GID of the project the task will be posted to;
            can also be provided to the `run` method
        - name (str, optional): the name of the task to create; can also be provided to the
            `run` method
        - notes (str, optional): the contents of the task; can also be provided to the `run` method
        - token (str): an Asana Personal Access Token
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task
            init method
    """

    def __init__(
        self,
        name: str = None,
        notes: str = None,
        project: str = None,
        token: str = None,
        **kwargs: Any
    ):
        self.name = name
        self.notes = notes
        self.project = project
        self.token = token
        super().__init__(**kwargs)

    @defaults_from_attrs("name", "notes", "project", "token")
    def run(
        self,
        name: str = None,
        notes: str = None,
        project: str = None,
        token: str = None,
    ) -> None:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a
        Flow context, or by using `Task.bind`.

        Args:
        - name (str, optional): the name of the task to create; can also be provided at initialization
        - project (str; , required): The GID of the project the task will be posted to;
            can also be provided at initialization
        - notes (str, optional): the contents of the task; can also be provided at initialization
        - token (str): an ASANA Personal Access Token

        Raises:
            - ValueError: if no project is provided
            - ValueError: if no access token is provided
            - ValueError: if no result is returned

        Returns:
            - The result object with details of the new asana task
        """

        if project is None:
            raise ValueError("An Asana project must be provided.")

        if token is None:
            raise ValueError("An Asana access token must be provided.")

        client = asana.Client.access_token(token)

        result = client.tasks.create_task(
            {"name": name, "notes": notes, "projects": [project]}
        )

        if not result:
            raise ValueError("Creating Asana Task failed")

        return result



import logging
import time
from typing import Dict, Iterable, Iterator, List, NamedTuple, Optional, Tuple, cast

import pendulum

import prefect
from prefect.backend.flow import FlowView
from prefect.backend.task_run import TaskRunView
from prefect.engine.state import State
from prefect.run_configs import RunConfig
from prefect.serialization.run_config import RunConfigSchema
from prefect.utilities.graphql import EnumValue, with_args
from prefect.utilities.logging import get_logger


logger = get_logger("backend.flow_run")


def stream_flow_run_logs(flow_run_id: str) -> None:
    """
    Basic wrapper for `watch_flow_run` to print the logs of the run

    EXPERIMENTAL: This interface is experimental and subject to change
    """
    for log in watch_flow_run(flow_run_id):
        level_name = logging.getLevelName(log.level)
        timestamp = log.timestamp.in_tz(tz="local")
        # Uses `print` instead of the logger to prevent duplicate timestamps
        print(
            f"{timestamp:%H:%M:%S} | {level_name:<7} | {log.message}",
        )


def watch_flow_run(
    flow_run_id: str,
    stream_states: bool = True,
    stream_logs: bool = True,
) -> Iterator["FlowRunLog"]:
    """
    Watch execution of a flow run displaying state changes. This function will yield
    `FlowRunLog` objects until the flow run enters a 'Finished' state.

    If both stream_states and stream_logs are `False` then this will just block until
    the flow run finishes.

    EXPERIMENTAL: This interface is experimental and subject to change

    Args:
        - flow_run_id: The flow run to watch
        - stream_states: If set, flow run state changes will be streamed as logs
        - stream_logs: If set, logs will be streamed from the flow run

    Yields:
        FlowRunLog: Sorted log entries

    """

    flow_run = FlowRunView.from_flow_run_id(flow_run_id)

    if flow_run.state.is_finished():
        time_ago = flow_run.updated_at.diff_for_humans()
        if stream_states:
            yield FlowRunLog(
                timestamp=pendulum.now(),
                level=logging.INFO,
                message=f"Your flow run finished {time_ago}",
            )
        return

    # The timestamp of the last displayed log so that we can scope each log query
    # to logs that have not been shown yet
    last_log_timestamp = None

    # A counter of states that have been displayed. Not a set so repeated states in the
    # backend are shown well.
    seen_states = 0

    # Some times (in seconds) to track displaying a warning about the flow not starting
    # on time
    agent_warning_initial_wait = 15
    agent_warning_repeat_interval = 30
    total_time_elapsed = 0
    agent_warning_time_elapsed = agent_warning_repeat_interval  # show first warning

    # We'll do a basic backoff for polling, not exposed as args because the user
    # probably should not need to tweak these.
    poll_min = poll_interval = 2
    poll_max = 10
    poll_factor = 1.3

    while not flow_run.state.is_finished():
        # Get the latest state
        flow_run = flow_run.get_latest()

        # Get a rounded time elapsed for display purposes
        total_time_elapsed_rounded = round(total_time_elapsed / 5) * 5
        # Check for a really long run
        if total_time_elapsed > 60 * 60 * 12:
            raise RuntimeError(
                "`watch_flow_run` timed out after 12 hours of waiting for completion. "
                "Your flow run is still in state: {flow_run.state}"
            )

        if (
            stream_states  # The agent warning is counted as a state log
            and total_time_elapsed >= agent_warning_initial_wait
            and agent_warning_time_elapsed > agent_warning_repeat_interval
            and not (
                flow_run.state.is_submitted()
                or flow_run.state.is_running()
                or flow_run.state.is_finished()
            )
        ):
            agent_msg = check_for_compatible_agents(flow_run.labels)
            yield FlowRunLog(
                timestamp=pendulum.now(),
                level=logging.WARN,
                message=(
                    f"It has been {total_time_elapsed_rounded} seconds and "
                    f"your flow run has not been submitted by an agent. {agent_msg}"
                ),
            )
            agent_warning_time_elapsed = 0

        # Create a shared message list so we can roll state changes and logs together
        # in the correct order
        messages = []

        # Add state change messages
        if stream_states:
            for state in flow_run.states[seen_states:]:
                state_name = type(state).__name__
                messages.append(
                    # Create some fake run logs with the state transitions
                    FlowRunLog(
                        timestamp=state.timestamp,  # type: ignore
                        level=logging.INFO,
                        message=f"Entered state <{state_name}>: {state.message}",
                    )
                )
                seen_states += 1

        if stream_logs:
            # Display logs if asked and the flow is running
            logs = flow_run.get_logs(
                start_time=last_log_timestamp, end_time=pendulum.now()
            )
            messages += logs
            if logs:
                # Set the last timestamp so the next query is scoped to logs we have
                # not seen yet`
                last_log_timestamp = logs[-1].timestamp

        for flow_run_log in sorted(messages):
            yield flow_run_log

        if not messages:
            # Delay the poll if there are no messages
            poll_interval = int(poll_interval * poll_factor)
        else:
            # Otherwise reset to the min poll time for a fast query
            poll_interval = poll_min

        poll_interval = min(poll_interval, poll_max)
        time.sleep(poll_interval)
        agent_warning_time_elapsed += poll_interval
        total_time_elapsed += poll_interval


def check_for_compatible_agents(labels: Iterable[str], since_minutes: int = 1) -> str:
    """
    Checks for agents compatible with a set of labels returning a user-friendly message
    indicating the status, roughly one of the following cases:

    - There is a healthy agent with matching labels
    - There are N healthy agents with matching labels
    - There is an unhealthy agent with matching labels but no healthy agents matching
    - There are N unhealthy agents with matching labels but no healthy agents matching
    - There are no healthy agents at all and no unhealthy agents with matching labels
    - There are healthy agents but no healthy or unhealthy agent has matching labels

    EXPERIMENTAL: This interface is experimental and subject to change

    Args:
        - labels: A set of labels; typically associated with a flow run
        - since_minutes: The amount of time in minutes to allow an agent to be idle and
            considered active/healthy still

    Returns:
        A message string
    """
    client = prefect.Client()

    labels = set(labels)
    labels_blurb = f"labels {labels!r}" if labels else "empty labels"

    result = client.graphql(
        {"query": {"agent": {"last_queried", "labels", "name", "id"}}}
    )

    agents = result.get("data", {}).get("agent")
    if agents is None:
        raise ValueError(f"Received bad result while querying for agents: {result}")

    # Parse last query times
    for agent in agents:
        agent.last_queried = cast(
            Optional[pendulum.DateTime],
            pendulum.parse(agent.last_queried) if agent.last_queried else None,
        )

    # Drop agents that have not queried
    agents = [agent for agent in agents if agent.last_queried is not None]

    # Drop agents that have not sent a recent hearbeat
    since = pendulum.now().subtract(minutes=since_minutes)
    healthy_agents = [agent for agent in agents if agent.last_queried >= since]

    # Search for the flow run labels in running agents
    matching_healthy = []
    matching_unhealthy = []

    for agent in agents:
        empty_labels_match = not agent.labels and not labels
        if empty_labels_match or (labels and labels.issubset(agent.labels)):
            if agent in healthy_agents:
                matching_healthy.append(agent)
            else:
                matching_unhealthy.append(agent)

    if len(matching_healthy) == 1:
        agent = matching_healthy[0]
        # Display the single matching agent
        name_blurb = f" ({agent.name})" if agent.name else ""
        return (
            f"Agent {agent.id}{name_blurb} has matching labels and last queried "
            f"{agent.last_queried.diff_for_humans()}. It should deploy your flow run."
        )

    if len(matching_healthy) > 1:
        # Display that there are multiple matching agents
        return (
            f"Found {len(matching_healthy)} healthy agents with matching labels. One "
            "of them should pick up your flow."
        )

    # We now know we have no matching healthy agents...

    if not healthy_agents and not matching_unhealthy:
        # Display that there are no matching agents all-time
        return (
            "There are no healthy agents in your tenant and it does not look like an "
            "agent with the required labels has been run recently. Start an agent with "
            f"{labels_blurb} to run your flow."
        )

    if len(matching_unhealthy) == 1:
        agent = matching_unhealthy[0]
        # Display that there is a single matching unhealthy agent
        name_blurb = f" ({agent.name})" if agent.name else ""
        return (
            f"Agent {agent.id}{name_blurb} has matching labels and last queried "
            f"{agent.last_queried.diff_for_humans()}. Since it hasn't queried recently, it looks "
            f"unhealthy. Restart it or start a new agent with {labels_blurb} to deploy "
            f"your flow run."
        )

    if len(matching_unhealthy) > 1:
        # Display that there are multiple matching unhealthy agents
        return (
            f"Found {len(matching_unhealthy)} agents with matching labels but they "
            "have not queried recently and look unhealthy. Restart one of them or "
            f"start a new agent with {labels_blurb} deploy your flow run."
        )

    # No matching healthy or unhealthy agents
    return (
        f"You have {len(healthy_agents)} healthy agents in your tenant but do not have "
        f"an agent with {labels_blurb}. Start an agent with matching labels to deploy "
        "your flow run."
    )


class FlowRunLog(NamedTuple):
    """
    Small wrapper for backend log objects

    EXPERIMENTAL: This interface is experimental and subject to change
    """

    timestamp: pendulum.DateTime
    level: int
    message: str

    @classmethod
    def from_dict(cls, data: dict) -> "FlowRunLog":
        return cls(
            cast(pendulum.DateTime, pendulum.parse(data["timestamp"])),
            logging.getLevelName(data["level"]),  # actually gets level int from name
            data["message"],
        )


class _TimestampedState(State):
    """
    Small wrapper for flow run states to include a timestamp

    TODO: We will likely want to include timestamps directly on `State`. This extension
          is written as a subclass for compatibility with that future change
    """

    timestamp: pendulum.DateTime

    @classmethod
    def from_dict(cls, data: dict) -> "_TimestampedState":
        state = cls.deserialize(data["serialized_state"])
        state = cast(_TimestampedState, state)
        # Our 3.6 compatible version of pendulum does not have `fromisoformat` so we
        # parse and cast
        state.timestamp = cast(pendulum.DateTime, pendulum.parse(data["timestamp"]))
        return state


class FlowRunView:
    """
    A view of Flow Run data stored in the Prefect API.

    Provides lazy loading of Task Runs from the flow run.

    This object is designed to be an immutable view of the data stored in the Prefect
    backend API at the time it is created. However, each time a task run is retrieved
    the latest data for that task will be pulled since they are loaded lazily. Finished
    task runs will be cached in this object to reduce the amount of network IO.

    EXPERIMENTAL: This interface is experimental and subject to change

    Args:
        - flow_run_id: The uuid of the flow run
        - name: The name of the flow run
        - flow_id: The uuid of the flow this run is associated with
        - state: The state of the flow run
        - labels: The labels assigned to this flow run
        - parameters: Parameter overrides for this flow run
        - context: Context overrides for this flow run
        - updated_at: When this flow run was last updated in the backend
        - run_config: The `RunConfig` this flow run was configured with
        - states: A sorted list of past states the flow run has been in
        - task_runs: An iterable of task run metadata to cache in this view
    """

    def __init__(
        self,
        flow_run_id: str,
        name: str,
        flow_id: str,
        labels: List[str],
        parameters: dict,
        context: dict,
        state: State,
        states: List[_TimestampedState],
        updated_at: pendulum.DateTime,
        run_config: "RunConfig",
        task_runs: Iterable["TaskRunView"] = None,
    ):
        self.flow_run_id = flow_run_id
        self.name = name
        self.flow_id = flow_id
        self.state = state
        self.labels = labels
        self.parameters = parameters
        self.context = context
        self.updated_at = updated_at
        self.states = states
        self.run_config = run_config

        # Cached value of all task run ids for this flow run, only cached if the flow
        # is done running
        self._task_run_ids: Optional[List[str]] = None

        # Cached value of flow metadata
        self._flow: Optional[FlowView] = None

        # Store a mapping of task run ids to task runs
        self._cached_task_runs: Dict[str, "TaskRunView"] = {}

        # Store a mapping of (slug, map_index) to task run ids
        self._slug_index_to_cached_id: Dict[Tuple[str, int], str] = {}

        if task_runs is not None:
            for task_run in task_runs:
                self._cache_task_run_if_finished(task_run)

    def _cache_task_run_if_finished(self, task_run: "TaskRunView") -> None:
        """
        Add a task run to the cache if it is in a finished state

        Args:
            - task_run: The task run to add
        """
        if task_run.state.is_finished():
            self._cached_task_runs[task_run.task_run_id] = task_run
            self._slug_index_to_cached_id[
                (task_run.task_slug, task_run.map_index)
            ] = task_run.task_run_id

    def get_latest(self, load_static_tasks: bool = False) -> "FlowRunView":
        """
        Get the a new copy of this object with the latest data from the API. Cached
        `TaskRunView` objects will be passed to the new object. Only finished tasks
        are cached so the cached data cannot be stale.

        This will not mutate the current object.

        Args:
            - load_static_tasks: Pre-populate the task runs with results from flow tasks
                that are unmapped. Defaults to `False` because it may be wasteful to
                query for task run data when cached tasks are already copied over
                from the old object.

        Returns:
            A new instance of FlowRunView
        """
        return self.from_flow_run_id(
            flow_run_id=self.flow_run_id,
            load_static_tasks=load_static_tasks,
            _cached_task_runs=self._cached_task_runs.values(),
        )

    def get_logs(
        self,
        start_time: pendulum.DateTime = None,
        end_time: pendulum.DateTime = None,
    ) -> List["FlowRunLog"]:
        """
        Get logs for this flow run from `start_time` to `end_time`.

        Args:
            - start_time (optional): A time to start the log query at, useful for
                limiting the scope. If not provided, all logs up to `updated_at` are
                retrieved.
            - end_time (optional): A time to end the log query at. By default, this is
                set to `self.updated_at` which is the last time that the flow run was
                updated in the backend before this object was created.

        Returns:
            A list of `FlowRunLog` objects sorted by timestamp
        """

        client = prefect.Client()
        end_time = end_time or self.updated_at

        logs_query = {
            with_args(
                "logs",
                {
                    "order_by": {EnumValue("timestamp"): EnumValue("asc")},
                    "where": {
                        "_and": [
                            {"timestamp": {"_lte": end_time.isoformat()}},
                            (
                                {"timestamp": {"_gt": start_time.isoformat()}}
                                if start_time
                                else {}
                            ),
                        ]
                    },
                },
            ): {"timestamp": True, "message": True, "level": True}
        }

        result = client.graphql(
            {
                "query": {
                    with_args(
                        "flow_run",
                        {
                            "where": {"id": {"_eq": self.flow_run_id}},
                        },
                    ): logs_query
                }
            }
        )

        # Unpack the result
        logs = result.get("data", {}).get("flow_run", [{}])[0].get("logs", [])

        return [FlowRunLog.from_dict(log) for log in logs]

    def get_flow_metadata(self, no_cache: bool = False) -> "FlowView":
        """
        Flow metadata for the flow associated with this flow run. Retrieved from the
        API on first call then cached for future calls.

        Args:
            - no_cache: If set, the cached `FlowView` will be ignored and the latest
                data for the flow will be pulled.

        Returns:
            FlowView: A view of the Flow metadata for the Flow this run is from
        """
        if self._flow is None or no_cache:
            self._flow = FlowView.from_flow_id(flow_id=self.flow_id)

        return self._flow

    @classmethod
    def _from_flow_run_data(
        cls, flow_run_data: dict, task_runs: Iterable["TaskRunView"] = None
    ) -> "FlowRunView":
        """
        Instantiate a `TaskRunView` from serialized data.

        This method deserializes objects into their Prefect types.

        Exists to maintain consistency in the design of backend "View" classes.

        Args:
            - flow_run_data: A dict of flow run data
            - task_runs: An optional iterable of task runs to pre-populate the cache with

        Returns:
            A populated `FlowRunView` instance
        """
        flow_run_data = flow_run_data.copy()  # Avoid mutating the input object

        flow_run_id = flow_run_data.pop("id")
        state = State.deserialize(flow_run_data.pop("serialized_state"))
        run_config = RunConfigSchema().load(flow_run_data.pop("run_config"))

        states_data = flow_run_data.pop("states", [])
        states = list(
            sorted(
                [_TimestampedState.from_dict(state_data) for state_data in states_data],
                key=lambda s: s.timestamp,
            )
        )
        updated_at = cast(
            pendulum.DateTime, pendulum.parse(flow_run_data.pop("updated"))
        )

        return cls(
            flow_run_id=flow_run_id,
            task_runs=task_runs,
            state=state,
            updated_at=updated_at,
            states=states,
            run_config=run_config,
            **flow_run_data,
        )

    @classmethod
    def from_flow_run_id(
        cls,
        flow_run_id: str,
        load_static_tasks: bool = False,
        _cached_task_runs: Iterable["TaskRunView"] = None,
    ) -> "FlowRunView":
        """
        Get an instance of this class filled with information by querying for the given
        flow run id

        Args:
            - flow_run_id: the flow run id to lookup
            - load_static_tasks: Pre-populate the task runs with results from flow tasks
                that are unmapped.
            - _cached_task_runs: Pre-populate the task runs with an existing iterable of
                task runs

        Returns:
            A populated `FlowRunView` instance
        """
        flow_run_data = cls._query_for_flow_run(where={"id": {"_eq": flow_run_id}})

        if load_static_tasks:
            task_run_data = TaskRunView._query_for_task_runs(
                where={
                    "map_index": {"_eq": -1},
                    "flow_run_id": {"_eq": flow_run_id},
                },
            )
            task_runs = [
                TaskRunView._from_task_run_data(data) for data in task_run_data
            ]

        else:
            task_runs = []

        # Combine with the provided `_cached_task_runs` iterable
        task_runs = task_runs + list(_cached_task_runs or [])

        return cls._from_flow_run_data(flow_run_data, task_runs=task_runs)

    @staticmethod
    def _query_for_flow_run(where: dict) -> dict:
        client = prefect.Client()

        flow_run_query = {
            "query": {
                with_args("flow_run", {"where": where}): {
                    "id": True,
                    "name": True,
                    "flow_id": True,
                    "serialized_state": True,
                    "states": {"timestamp", "serialized_state"},
                    "labels": True,
                    "parameters": True,
                    "context": True,
                    "updated": True,
                    "run_config": True,
                }
            }
        }

        result = client.graphql(flow_run_query)
        flow_runs = result.get("data", {}).get("flow_run", None)

        if flow_runs is None:
            raise ValueError(
                f"Received bad result while querying for flow runs where {where}: "
                f"{result}"
            )

        if not flow_runs:
            raise ValueError(
                f"No flow runs found while querying for flow runs where {where}"
            )

        if len(flow_runs) > 1:
            raise ValueError(
                f"Found multiple ({len(flow_runs)}) flow runs while querying for flow "
                f"runs where {where}: {flow_runs}"
            )

        return flow_runs[0]

    def get_task_run(
        self,
        task_slug: str = None,
        task_run_id: str = None,
        map_index: int = None,
    ) -> "TaskRunView":
        """
        Get information about a task run from this flow run. Lookup is available by one
        of the arguments. If the task information is not available locally already,
        we will query the database for it. If multiple arguments are provided, we will
        validate that they are consistent with each other.

        All retrieved task runs that are finished will be cached to avoid re-querying in
        repeated calls

        Args:
            - task_slug: A task slug string to use for the lookup
            - task_run_id: A task run uuid to use for the lookup
            - map_index: If given a slug of a mapped task, an index may be provided to
                get the the task run for that child task instead of the parent. This
                value will only be used for a consistency check if passed with a
                `task_run_id`

        Returns:
            A cached or newly constructed TaskRunView instance
        """

        if task_run_id is not None:
            # Load from the cache if available or query for results
            result = (
                self._cached_task_runs[task_run_id]
                if task_run_id in self._cached_task_runs
                else TaskRunView.from_task_run_id(task_run_id)
            )

            if task_slug is not None and result.task_slug != task_slug:
                raise ValueError(
                    "Both `task_slug` and `task_run_id` were provided but the task "
                    "found using `task_run_id` has a different slug! "
                    f"`task_slug == {task_slug!r}` and "
                    f"`result.task_slug == {result.task_slug!r}`"
                )

            if map_index is not None and result.map_index != map_index:
                raise ValueError(
                    "Both `map_index` and `task_run_id` were provided but the task "
                    "found using `task_run_id` has a different map index! "
                    f"`map_index == {map_index}` and "
                    f"`result.map_index == {result.map_index}`"
                )

            self._cache_task_run_if_finished(result)
            return result

        if task_slug is not None:

            # Default to loading the parent task
            if map_index is None:
                map_index = -1

            # Check the cache
            task_run_id = self._slug_index_to_cached_id.get((task_slug, map_index))
            if task_run_id:
                return self._cached_task_runs[task_run_id]

            result = TaskRunView.from_task_slug(
                task_slug=task_slug, flow_run_id=self.flow_run_id, map_index=map_index
            )
            self._cache_task_run_if_finished(result)
            return result

        raise ValueError(
            "One of `task_run_id`, `task`, or `task_slug` must be provided!"
        )

    def get_all_task_runs(self) -> List["TaskRunView"]:
        """
        Get all task runs for this flow run in a single query. Finished task run data
        is cached so future lookups do not query the backend.

        Returns:
            A list of TaskRunView objects
        """
        if len(self.get_task_run_ids()) > 1000:
            raise ValueError(
                "Refusing to `get_all_task_runs` for a flow with more than 1000 tasks. "
                "Please load the tasks you are interested in individually."
            )

        # Run a single query instead of querying for each task run separately
        task_run_data = TaskRunView._query_for_task_runs(
            where={
                "flow_run_id": {"_eq": self.flow_run_id},
                "id": {"_nin": list(self._cached_task_runs.keys())},
            }
        )

        new_task_runs = [
            TaskRunView._from_task_run_data(data) for data in task_run_data
        ]

        task_runs = new_task_runs + list(self._cached_task_runs.values())

        for task_run in new_task_runs:
            self._cache_task_run_if_finished(task_run)

        return task_runs

    def get_task_run_ids(self) -> List[str]:
        """
        Get all task run ids associated with this flow run. Lazily loaded at call time
        then cached for future calls.

        Returns:
            A list of string task run ids
        """
        # Return the cached value immediately if it exists
        if self._task_run_ids:
            return self._task_run_ids

        client = prefect.Client()

        task_query = {
            "query": {
                with_args(
                    "task_run",
                    {
                        "where": {
                            "flow_run_id": {"_eq": self.flow_run_id},
                        }
                    },
                ): {
                    "id": True,
                }
            }
        }
        result = client.graphql(task_query)
        task_runs = result.get("data", {}).get("task_run", None)

        if task_runs is None:
            logger.warning(
                f"Failed to load task run ids for flow run {self.flow_run_id}: "
                f"{result}"
            )

        task_run_ids = [task_run["id"] for task_run in task_runs]

        # If the flow run is done, we can safely cache this value
        if self.state.is_finished():
            self._task_run_ids = task_run_ids

        return task_run_ids

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}"
            "("
            + ", ".join(
                [
                    f"flow_run_id={self.flow_run_id!r}",
                    f"name={self.name!r}",
                    f"state={self.state!r}",
                    f"labels={self.labels!r}",
                    f"cached_task_runs={len(self._cached_task_runs)}",
                ]
            )
            + ")"
        )



"""
A collection of tasks for running Jupyter notebooks.
"""
try:
    from prefect.tasks.jupyter.jupyter import ExecuteNotebook
except ImportError as import_error:
    raise ImportError(
        'Using `prefect.tasks.jupyter` requires Prefect to be installed with the "jupyter" extra.'
    ) from import_error



"""
The tasks in this module can be used to represent arbitrary functions.

In general, users will not instantiate these tasks by hand; they will
automatically be applied when users apply the `@task` decorator.
"""

from typing import Any, Callable
from functools import update_wrapper

import prefect


class FunctionTask(prefect.Task):
    """A convenience Task for functionally creating Task instances with
    arbitrary callable `run` methods.

    Args:
        - fn (callable): the function to be the task's `run` method
        - name (str, optional): the name of this task; if not provided it is inferred
            as the function name
        - **kwargs: keyword arguments that will be passed to the Task
            constructor

    Raises:
        - ValueError: if the provided function violates signature requirements
            for Task run methods

    Example:
    ```python
    task = FunctionTask(lambda x: x - 42, name="Subtract 42")

    with Flow("My Flow") as f:
        result = task(42)
    ```
    """

    def __init__(self, fn: Callable, name: str = None, **kwargs: Any):
        if not callable(fn):
            raise TypeError("`fn` must be callable")

        # Set the Prefect name from the function
        if name is None:
            name = getattr(fn, "__name__", type(self).__name__)

        prefect.core.task._validate_run_signature(fn)
        self.run = fn
        update_wrapper(self, fn)

        super().__init__(name=name, **kwargs)



"""
Functions that exist for compatibility with different python versions. These should
only be used internally.
"""

import sys

# Provide `nullcontext`

if sys.version_info < (3, 7):

    from contextlib import contextmanager
    from typing import Iterator

    @contextmanager
    def nullcontext() -> Iterator[None]:
        yield


else:

    from contextlib import nullcontext  # noqa: F401



import os
from typing import TYPE_CHECKING, Any
from urllib.parse import quote_plus

import prefect
from prefect.client import Secret
from prefect.storage import Storage
from prefect.utilities.storage import extract_flow_from_file

if TYPE_CHECKING:
    from prefect.core.flow import Flow
    from gitlab import Gitlab


class GitLab(Storage):
    """
    GitLab storage class. This class represents the Storage interface for Flows stored
    in `.py` files in a GitLab repository.

    This class represents a mapping of flow name to file paths contained in the git repo,
    meaning that all flow files should be pushed independently. A typical workflow using
    this storage type might look like the following:

    - Compose flow `.py` file where flow has GitLab storage:

    ```python
    flow = Flow("my-flow")
    # Can also use `repo="123456"`
    flow.storage = GitLab(repo="my/repo", path="/flows/flow.py", ref="my-branch")
    ```

    - Push this `flow.py` file to the `my/repo` repository under `/flows/flow.py`.

    - Call `prefect register flow -f flow.py` to register this flow with GitLab storage.

    Args:
        - repo (str): the project path (i.e., 'namespace/project') or ID
        - host (str, optional): If using GitLab server, the server host. If not
            specified, defaults to Gitlab cloud.
        - path (str, optional): a path pointing to a flow file in the repo
        - ref (str, optional): a commit SHA-1 value or branch name
        - access_token_secret (str, optional): The name of a Prefect secret
            that contains a GitLab access token to use when loading flows from
            this storage.
        - **kwargs (Any, optional): any additional `Storage` initialization options
    """

    def __init__(
        self,
        repo: str,
        host: str = None,
        path: str = None,
        ref: str = None,
        access_token_secret: str = None,
        **kwargs: Any,
    ) -> None:
        self.repo = repo
        self.host = host
        self.path = path
        self.ref = ref
        self.access_token_secret = access_token_secret

        super().__init__(**kwargs)

    def get_flow(self, flow_name: str) -> "Flow":
        """
        Given a flow name within this Storage object, load and return the Flow.

        Args:
            - flow_name (str): the name of the flow to return.

        Returns:
            - Flow: the requested flow
        """
        try:
            from gitlab.exceptions import GitlabAuthenticationError, GitlabGetError
        except ImportError as exc:
            raise ImportError(
                "Unable to import gitlab, please ensure you have installed the gitlab extra"
            ) from exc

        if flow_name not in self.flows:
            raise ValueError("Flow is not contained in this Storage")
        flow_location = self.flows[flow_name]

        ref = self.ref or "master"

        client = self._get_gitlab_client()

        try:
            project = client.projects.get(quote_plus(self.repo, safe="/"))
            contents = project.files.get(file_path=flow_location, ref=ref)
        except GitlabAuthenticationError:
            self.logger.error(
                "Unable to authenticate GitLab account. Please check your credentials."
            )
            raise
        except GitlabGetError:
            self.logger.error(
                f"Error retrieving file contents at {flow_location} in {self.repo}@{ref}. "
                "Ensure the project and file exist."
            )
            raise

        return extract_flow_from_file(
            file_contents=contents.decode(), flow_name=flow_name
        )

    def add_flow(self, flow: "Flow") -> str:
        """
        Method for storing a new flow as bytes in the local filesystem.

        Args:
            - flow (Flow): a Prefect Flow to add

        Returns:
            - str: the location of the added flow in the repo

        Raises:
            - ValueError: if a flow with the same name is already contained in this storage
        """
        if flow.name in self:
            raise ValueError(
                'Name conflict: Flow with the name "{}" is already present in this storage.'.format(
                    flow.name
                )
            )

        self.flows[flow.name] = self.path  # type: ignore
        self._flows[flow.name] = flow
        return self.path  # type: ignore

    def _get_gitlab_client(self) -> "Gitlab":
        from gitlab import Gitlab

        if self.access_token_secret is not None:
            # If access token secret specified, load it
            access_token = Secret(self.access_token_secret).get()
        else:
            # Otherwise, fallback to loading from local secret or environment variable
            access_token = prefect.context.get("secrets", {}).get("GITLAB_ACCESS_TOKEN")
            if access_token is None:
                access_token = os.getenv("GITLAB_ACCESS_TOKEN")

        host = "https://gitlab.com" if self.host is None else self.host

        return Gitlab(host, private_token=access_token)



from prefect.backend.task_run import TaskRunView
from prefect.backend.flow_run import FlowRunView
from prefect.backend.flow import FlowView
from prefect.backend.tenant import TenantView
from prefect.backend.kv_store import set_key_value, get_key_value, delete_key, list_keys
from prefect.backend.artifacts import (
    create_link_artifact,
    create_markdown_artifact,
    delete_artifact,
    update_link_artifact,
    update_markdown_artifact,
)



"""
Tasks for interacting with the Prefect API
"""

from prefect.tasks.prefect.flow_run import (
    create_flow_run,
    wait_for_flow_run,
    get_task_run_result,
)
from prefect.tasks.prefect.flow_run import StartFlowRun
from prefect.tasks.prefect.flow_run_rename import RenameFlowRun
from prefect.tasks.prefect.flow_run_cancel import CancelFlowRun



from typing import Any, Callable, Dict, List, Type, TYPE_CHECKING
from urllib.parse import urlparse

import prefect
from prefect import Client
from prefect.environments.execution import Environment

if TYPE_CHECKING:
    from prefect.core.flow import Flow
    from distributed.deploy.cluster import Cluster
    from distributed.security import Security


class DaskCloudProviderEnvironment(Environment):
    """
    DaskCloudProviderEnvironment creates Dask clusters using the Dask Cloud Provider
    project.

    DEPRECATED: Environment based configuration is deprecated, please transition to
    configuring `flow.run_config` instead of `flow.environment`. See
    https://docs.prefect.io/orchestration/flow_config/overview.html for more info.

    For each flow run, a new Dask cluster will be dynamically created and the
    flow will run using a `DaskExecutor` with the Dask scheduler address from the newly
    created Dask cluster. You can specify the number of Dask workers manually
    (for example, passing the kwarg `n_workers`) or enable adaptive mode by
    passing `adaptive_min_workers` and, optionally, `adaptive_max_workers`. This
    environment aims to provide a very easy path to Dask scalability for users of
    cloud platforms, like AWS.

    **NOTE:** AWS Fargate Task (not Prefect Task) startup time can be slow, depending
    on docker image size. Total startup time for a Dask scheduler and workers can
    be several minutes. This environment is a much better fit for production
    deployments of scheduled Flows where there's little sensitivity to startup
    time. `DaskCloudProviderEnvironment` is a particularly good fit for automated
    deployment of Flows in a CI/CD pipeline where the infrastructure for each Flow
    should be as independent as possible, e.g. each Flow could have its own docker
    image, dynamically create the Dask cluster to run on, etc. However, for
    development and interactive testing, creating a Dask cluster manually with Dask
    Cloud Provider and then using `LocalEnvironment` with a `DaskExecutor`
    will result in a much better development experience.

    (Dask Cloud Provider currently only supports AWS using either Fargate or ECS.
    Support for AzureML is coming soon.)

    *IMPORTANT* By default, Dask Cloud Provider may create a Dask cluster in some
    environments (e.g. Fargate) that is accessible via a public IP, without any
    authentication, and configured to NOT encrypt network traffic. Please be
    conscious of security issues if you test this environment. (Also see pull
    requests [85](https://github.com/dask/dask-cloudprovider/pull/85) and
    [91](https://github.com/dask/dask-cloudprovider/pull/91) in the Dask Cloud
    Provider project.)

    Args:
        - provider_class (class): Class of a provider from the Dask Cloud Provider
            projects. Current supported options are `ECSCluster` and `FargateCluster`.
        - adaptive_min_workers (int, optional): Minimum number of workers for adaptive
            mode. If this value is None, then adaptive mode will not be used and you
            should pass `n_workers` or the appropriate kwarg for the provider class you
            are using.
        - adaptive_max_workers (int, optional): Maximum number of workers for adaptive
            mode.
        - security (Type[Security], optional): a Dask Security object from
            `distributed.security.Security`.  Use this to connect to a Dask cluster that is
            enabled with TLS encryption.  For more on using TLS with Dask see
            https://distributed.dask.org/en/latest/tls.html
        - executor_kwargs (dict, optional): a dictionary of kwargs to be passed to
            the executor; defaults to an empty dictionary
        - labels (List[str], optional): a list of labels, which are arbitrary string
            identifiers used by Prefect Agents when polling for work
        - on_execute (Callable[[Dict[str, Any], Dict[str, Any]], None], optional): a function
            callback which will be called before the flow begins to run. The callback function
            can examine the Flow run parameters and modify  kwargs to be passed to the Dask
            Cloud Provider class's constructor prior to launching the Dask cluster for the Flow
            run.  This allows for dynamically sizing the cluster based on the Flow run
            parameters, e.g.  settings n_workers. The callback function's signature should be:
            `on_execute(parameters: Dict[str, Any], provider_kwargs: Dict[str, Any]) -> None`
            The callback function may modify provider_kwargs
            (e.g.  `provider_kwargs["n_workers"] = 3`) and any relevant changes will be used when
            creating the Dask cluster via a Dask Cloud Provider class.
        - on_start (Callable, optional): a function callback which will be called before the
            flow begins to run
        - on_exit (Callable, optional): a function callback which will be called after the flow
            finishes its run
        - metadata (dict, optional): extra metadata to be set and serialized on this environment
        - **kwargs (dict, optional): additional keyword arguments to pass to boto3 for
            `register_task_definition` and `run_task`
    """

    def __init__(  # type: ignore
        self,
        provider_class: "Type[Cluster]",
        adaptive_min_workers: int = None,
        adaptive_max_workers: int = None,
        security: "Security" = None,
        executor_kwargs: Dict[str, Any] = None,
        labels: List[str] = None,
        on_execute: Callable[[Dict[str, Any], Dict[str, Any]], None] = None,
        on_start: Callable = None,
        on_exit: Callable = None,
        metadata: dict = None,
        **kwargs
    ) -> None:
        self._provider_class = provider_class
        self._adaptive_min_workers = adaptive_min_workers
        self._adaptive_max_workers = adaptive_max_workers
        self._on_execute = on_execute
        self._provider_kwargs = kwargs
        self.executor_kwargs = (executor_kwargs or {}).copy()
        if "skip_cleanup" not in self._provider_kwargs:
            # Prefer this default (if not provided) to avoid deregistering task definitions See
            # this issue in Dask Cloud Provider:
            # https://github.com/dask/dask-cloudprovider/issues/94
            self._provider_kwargs["skip_cleanup"] = True
        self._security = security
        if self._security:
            # We'll use the security config object both for our Dask Client connection *and*
            # for the particular Dask Cloud Provider (e.g. Fargate) to use with *its* Dask
            # Client when it connects to the scheduler after cluster creation. So we
            # put it in _provider_kwargs so it gets passed to the Dask Cloud Provider's constructor
            self._provider_kwargs["security"] = self._security
            self.executor_kwargs["client_kwargs"] = {"security": self._security}

        self.cluster = None

        super().__init__(
            labels=labels,
            on_start=on_start,
            on_exit=on_exit,
            metadata=metadata,
        )

    @property
    def dependencies(self) -> list:
        return ["dask_cloudprovider"]

    def _create_dask_cluster(self) -> None:
        self.logger.info("Creating Dask cluster using {}".format(self._provider_class))
        self.cluster = self._provider_class(**self._provider_kwargs)
        if self.cluster and self.cluster.scheduler and self.cluster.scheduler.address:
            self.logger.info(
                "Dask cluster created. Scheduler address: {} Dashboard: http://{}:8787 "
                "(unless port was changed from default of 8787)".format(
                    self.cluster.scheduler.address,
                    urlparse(self.cluster.scheduler.address).hostname,
                )  # TODO submit PR to Dask Cloud Provider allowing discovery of dashboard port
            )

            self.executor_kwargs["address"] = self.cluster.scheduler.address  # type: ignore
        else:
            if self.cluster:
                self.cluster.close()
            raise Exception(
                "Unable to determine the Dask scheduler address after cluster creation. "
                "Tearting down cluster and terminating setup."
            )
        if self._adaptive_min_workers:
            self.logger.info(
                "Enabling adaptive mode with min_workers={} max_workers={}".format(
                    self._adaptive_min_workers, self._adaptive_max_workers
                )
            )
            self.cluster.adapt(  # type: ignore
                minimum=self._adaptive_min_workers, maximum=self._adaptive_max_workers
            )

    def execute(  # type: ignore
        self, flow: "Flow", **kwargs: Any  # type: ignore
    ) -> None:
        """
        Execute a flow run on a dask-cloudprovider cluster.

        Args:
            - flow (Flow): the Flow object
            - **kwargs (Any): Unused
        """
        flow_run_info = None
        flow_run_id = prefect.context.get("flow_run_id")
        if self._on_execute:
            # If an on_execute Callable has been provided, retrieve the flow run parameters
            # and then allow the Callable a chance to update _provider_kwargs. This allows
            # better sizing of the cluster resources based on parameters for this Flow run.
            try:
                client = Client()
                flow_run_info = client.get_flow_run_info(flow_run_id)
                parameters = flow_run_info.parameters or {}  # type: ignore
                self._on_execute(parameters, self._provider_kwargs)
            except Exception as exc:
                self.logger.info(
                    "Failed to retrieve flow run info with error: {}".format(repr(exc))
                )
        if "image" not in self._provider_kwargs or not self._provider_kwargs.get(
            "image"
        ):
            # If image is not specified, use the Flow's image so that dependencies are
            # identical on all containers: Flow runner, Dask scheduler, and Dask workers
            flow_id = prefect.context.get("flow_id")
            try:
                client = Client()
                if not flow_id:  # We've observed cases where flow_id is None
                    if not flow_run_info:
                        flow_run_info = client.get_flow_run_info(flow_run_id)
                    flow_id = flow_run_info.flow_id
                flow_info = client.graphql(
                    """query {
                  flow(where: {id: {_eq: "%s"}}) {
                    storage
                  }
                }"""
                    % flow_id
                )
                storage_info = flow_info["data"]["flow"][0]["storage"]
                image = "{}/{}:{}".format(
                    storage_info["registry_url"],
                    storage_info["image_name"],
                    storage_info["image_tag"],
                )
                self.logger.info(
                    "Using Flow's Docker image for Dask scheduler & workers: {}".format(
                        image
                    )
                )
                self._provider_kwargs["image"] = image
            except Exception as exc:
                self.logger.info(
                    "Failed to retrieve flow info with error: {}".format(repr(exc))
                )

        self._create_dask_cluster()

        self.logger.info(
            "Executing on dynamically created Dask Cluster with scheduler address: {}".format(
                self.executor_kwargs["address"]
            )
        )
        if self.on_start:
            self.on_start()

        try:
            from prefect.engine import get_default_flow_runner_class
            from prefect.executors import DaskExecutor

            runner_cls = get_default_flow_runner_class()
            runner_cls(flow=flow).run(executor=DaskExecutor(**self.executor_kwargs))
        except Exception as exc:
            self.logger.exception(
                "Unexpected error raised during flow run: {}".format(exc)
            )
            raise
        finally:
            if self.on_exit:
                self.on_exit()



import sys
import json
from typing import Any, List

import prefect
from prefect.utilities.graphql import with_args
from prefect.client import Client
from prefect.exceptions import ClientError


NON_CLOUD_BACKEND_ERROR_MESSAGE = (
    "KV Store operations are only supported while using Prefect Cloud as a backend."
)


def set_key_value(key: str, value: Any) -> str:
    """
    Set key value pair, overwriting values for existing key

    Args:
        - key (str): the name of the key
        - value (Any): A json compatible value

    Returns:
        - id (str): the id of the key value pair

    Raises:
        - ClientError: if using Prefect Server instead of Cloud
        - ValueError: if `value` exceeds 10 KB limit
    """
    if prefect.config.backend != "cloud":
        raise ClientError(NON_CLOUD_BACKEND_ERROR_MESSAGE)

    # check value is under size limit
    # note this will be enforced by the API
    value_size = sys.getsizeof(json.dumps(value))
    if value_size > 10000:  # 10 KB max
        raise ValueError("Value payload exceedes 10 KB limit.")

    mutation = {
        "mutation($input: set_key_value_input!)": {
            "set_key_value(input: $input)": {"id"}
        }
    }

    client = Client()
    result = client.graphql(
        query=mutation, variables=dict(input=dict(key=key, value=value))
    )

    return result.data.set_key_value.id


def get_key_value(key: str) -> Any:
    """
    Get the value for a key

    Args:
        - key (str): the name of the key

    Returns:
        - value (Any): A json compatible value

    Raises:
        - ValueError: if the specified key does not exist
        - ClientError: if using Prefect Server instead of Cloud
    """
    if prefect.config.backend != "cloud":
        raise ClientError(NON_CLOUD_BACKEND_ERROR_MESSAGE)

    query = {
        "query": {with_args("key_value", {"where": {"key": {"_eq": key}}}): {"value"}}
    }
    client = Client()
    result = client.graphql(query)  # type: Any
    if len(result.data.key_value) == 0:
        raise ValueError(f"No value found for key: {key}")
    return result.data.key_value[0].value


def delete_key(key: str) -> bool:
    """
    Delete a key value pair

    Args:
        - key (str): the name of the key

    Returns:
        - success (bool): Whether or not deleting the key succeeded

    Raises:
        - ValueError: if the specified key does not exist
        - ClientError: if using Prefect Server instead of Cloud
    """
    if prefect.config.backend != "cloud":
        raise ClientError(NON_CLOUD_BACKEND_ERROR_MESSAGE)

    query = {
        "query": {with_args("key_value", {"where": {"key": {"_eq": key}}}): {"id"}}
    }
    mutation = {
        "mutation($input: delete_key_value_input!)": {
            "delete_key_value(input: $input)": {"success"}
        }
    }

    client = Client()
    key_value_id_query = client.graphql(query=query)
    if len(key_value_id_query.data.key_value) == 0:
        raise ValueError(f"No key {key} found to delete")
    result = client.graphql(
        query=mutation,
        variables=dict(
            input=dict(key_value_id=key_value_id_query.data.key_value[0].id)
        ),
    )

    return result.data.delete_key_value.success


def list_keys() -> List[str]:
    """
    List all keys

    Returns:
        - keys (list): A list of keys

    Raises:
        - ClientError: if using Prefect Server instead of Cloud
    """
    if prefect.config.backend != "cloud":
        raise ClientError(NON_CLOUD_BACKEND_ERROR_MESSAGE)
    client = Client()
    result = client.graphql({"query": {"key_value": {"key"}}})  # type: ignore
    return sorted([res["key"] for res in result.data.key_value])



"""
This module contains a collection of tasks to run Data Quality tests using soda-spark library
"""

try:
    from prefect.tasks.sodaspark.sodaspark_tasks import SodaSparkScan
except ImportError as err:
    raise ImportError(
        'Using `prefect.tasks.sodaspark` requires Prefect to be installed with the "sodaspark" extra.'
    ) from err
